/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include <math.h>
#include <sys/types.h>
#include <signal.h>

#include "options.h"
#include "statistics.h"
#include "concurency.h"
#include "ALC.h"

void ALC::init()
{
  Formula::computeWeight = computeWeight;
  allocChildInfo();

  Options::init();
  FormulaCatalogue::init();

  InOut::inout = new ALCInOut();
  InOut::inout->init();
};

void ALC::initChild(int seconds)
{
  Graph::graph = new ALCGraph();
  STATISTICS_START(totalTime);
  scheduleExitTimeForChild(seconds);
};

char ALC::checkSatisfiability(Formula* checkedFormula, Formula* generalAssumptions, 
                              int seconds, bool clearParentFCat)
{
  // assume that ALC::init() has been called

  int pid = fork();
  if (pid < 0) {
    MSG("Error on fork!");
    deallocChildInfo();
    exit(-1);
  };

  if (pid == 0) { // the child process

    if (Options::weightingMethod == WM_DIAMOND_DEPTH) {
      if (checkedFormula->getWeight() < 10 && generalAssumptions->getWeight() < 10) {
        Options::usingSize4OrBranching = true;
        Options::usingSize4SelectingDfsBranch = true;
      };
    };

    ALC::initChild(seconds);
    Graph *graph = Graph::graph;
    graph->init(checkedFormula, generalAssumptions);

    if (graph->checkSatisfiability()) {
      DEBUG_MSG("Child exits with result SAT.");
      STATISTICS_SET(result, SAT);
    }
    else {
      DEBUG_MSG("Child exits with result UNSAT.");
      STATISTICS_SET(result, UNSAT);
    };

    STATISTICS_STOP(totalTime);
    STATISTICS_SET(exited, true);
    exit(EXIT_SUCCESS);
  }
  else { // the parent process
    if (clearParentFCat) {
      FormulaCatalogue::clear();
      FormulaCatalogue::init();
    };
    for (int i = 0; i < seconds + 2; i++) {
      if (_stat->exited) break;
      sleep(1);
    };
    if (!_stat->exited) {
      kill(pid, SIGTERM);
      for (int i = 0; i < Options::secondsBeforeSIGKILL; i++) {
        if (_stat->exited) break;
        sleep(1);
      };
      if (!_stat->exited) {
        STATISTICS_STOP(totalTime);
        STATISTICS_SET(exited, true);
        kill(pid, SIGKILL);
        sleep(1);
      };
    };
    assert(_stat->exited);
    return _stat->result;
  };
};

#define display_statistics() { \
    cout << "numOfExpansions: " << totalStat.numOfExpansions << "\n"; \
    cout << "applyingRules: " << totalStat.applyingRules << "s\n"; \
    cout << "choosingRules: " << totalStat.choosingRules << "s\n"; \
    if (Options::useUnsatCores) { \
      cout << "computingUnsatCores: " << totalStat.computingUnsatCores << "s\n"; \
      cout << "usingUnsatCores: " << totalStat.usingUnsatCores << "s\n"; \
    }; \
    if (Options::subsetChecking) \
      cout << "subsetChecking: " << totalStat.subsetChecking << "s\n"; \
    if (Options::subsetChecking1) \
      cout << "subsetChecking1: " << totalStat.subsetChecking1 << "s\n"; \
    if (Options::subsetChecking2) \
      cout << "subsetChecking2: " << totalStat.subsetChecking2 << "s\n"; \
    if (Options::localSupersetChecking) \
      cout << "localSupersetChecking: " << totalStat.localSupersetChecking << "s\n"; \
    if (Options::subsetChecking || Options::subsetChecking2) \
      cout << "updatingUnsats: " << totalStat.updatingUnsats << "s\n"; \
    if (Options::literalElimination) \
      cout << "eliminatingLiterals: " << totalStat.eliminatingLiterals << "s\n"; \
    if (Options::semanticBranching1) \
      cout << "semanticBranching1: " << totalStat.semanticBranching1 << "s\n"; \
    if (Options::semanticBranching2) \
      cout << "semanticBranching2: " << totalStat.semanticBranching2 << "s\n"; \
    if (totalStat.detectingSatNodes > 0) \
      cout << "detectingSatNodes: " << totalStat.detectingSatNodes << "s\n"; \
    if (totalStat.recomputingExpandingQueue > 0) \
      cout << "recomputingExpandingQueue: " << totalStat.recomputingExpandingQueue << "s\n"; \
    if (Options::hsExpanding) \
      cout << "computingNodePriorities: " << totalStat.computingNodePriorities << "s\n"; \
    if (Options::cutoff) \
      cout << "updatingReachabilityRelation: " << totalStat.updatingReachabilityRelation << "s\n"; \
    if (totalStat.mergingNodes > 0) \
      cout << "mergingNodes: " << totalStat.mergingNodes << "s\n"; \
    cout << "compactingFormulaCatalogue: " << totalStat.compactingFormulaCatalogue << "s\n";\
    if (Options::simplifyFormulas) \
      cout << "simplifyingFormulas: " << totalStat.simplifyingFormulas << "s\n"; \
}

void ALC::test_batch_file(int command, const char *fileName, int seconds)
{
    FILE *fi;
    if (!(fi = fopen(fileName, "r"))) {
      cout << "Error on opening the input file." << endl;
      exit(-1);
    };

    clock_t cl1 = clock();
    time_t t1, t2;
    time(&t1);

    ALC::init();
    ALCInOut* io = (ALCInOut*) InOut::inout;

    Options::minExpansionsBDSN = 1000000; // avoid detecting looping SAT nodes
    Options::minFormulasBeforeCompactingCatalogue = 30000; 
    Options::expectedCatalogueCompactingGain = 0.40;

    if (!Options::usingOptionsFile) {
      Options::hsExpanding  = false;
      Options::dfsExpanding = true;
      Options::breakWhenUnsuccessful = (command == 2);
    };

    Statistics totalStat;

    io->omitSeperators(fi);
    while (!feof(fi)) {
      STATISTICS_INIT();

      Formula* f;
      if (command == 1) // T98-sat
        f = io->getFormula(fi);
      else { // T98-tbox
        io->readAcyclicTBox(fi);
        f = io->getTESTformula();
        io->clearConceptDefs();
      };
      io->omitSeperators(fi);

      char rs = checkSatisfiability(f, FormulaCatalogue::top, seconds, true);

      if (Options::verbose)
        cout << (rs == SAT ? "SAT" : (rs == UNSAT ? "UNSAT" : "UNKNOWN"))
             << ": " << ((double) _stat->totalTime[0]) / CLOCKS_PER_SEC << "s\n";

      totalStat.add(_stat, Options::breakWhenUnsuccessful);

      if ((rs == UNKNOWN) && Options::breakWhenUnsuccessful)
        break;
    };
    fclose(fi);

    cout << "\nStatistics:\n\n";
    cout << "SAT:     " << totalStat.numOfSatResults << "\n";
    cout << "UNSAT:   " << totalStat.numOfUnsatResults << "\n";
    cout << "UNKNOWN: " << totalStat.numOfUnknownResults << "\n";

    totalStat.totalTime += ((double) clock() - cl1) / CLOCKS_PER_SEC;
    time(&t2);
    if (Options::breakWhenUnsuccessful) {
      cout << "\nTotal CPU time for the solved problems: " << totalStat.totalTime << "s\n";
      cout << "\nTotal time: " << t2 - t1 + 1 
	<< "s (CPU time + sleeping time + time for unsolved problems)\n\n";
    }
    else {
      cout << "\nTotal CPU time: " << totalStat.totalTime << "s\n";
      cout << "\nTotal time: " << t2 - t1 + 1 
	<< "s (CPU time + sleeping time)\n\n";
    };

    if (Options::verbose)
      display_statistics();

    deallocChildInfo();
    delete io;
    exit(EXIT_SUCCESS);
}

void ALC::main()
{
  MSG("This is version 1.2 of TGC (Tableaux with Global Caching).");
  MSG("It supports only checking satisfiability of a concept w.r.t. a TBox in ALC.\n");
  MSG("Available commands: ");
  MSG("  1. Test a batch file in the format T98-sat.");
  MSG("  2. Test a batch file in the format T98-kb.");
  MSG("  3. Generate a random TBox.");
  MSG("  4. Load a TBox.");
  MSG("  5. Test iteractively using the loaded TBox.");
  MSG("  6. Test randomly using the loaded TBox.");
  cout << "\nPlease choose a command: ";

  int command;
  while (true) {
    cin >> command;
    if (command >= 1 && command <= 6)
      break;
    else
      cout << "Enter a natural number from 1 to 6: ";
  };

  //---------------------------------------------------------------------------

  if (command == 1 || command == 2) {

    string fileName;
    cout << "Enter the path to the input file: ";
    cin >> fileName;

    int seconds;
    cout << "Enter the number of seconds used for each single test: ";
    cin >> seconds;

    test_batch_file(command, fileName.c_str(), seconds);
  }; // if

  //---------------------------------------------------------------------------

  ALC::init();
  ALCInOut* io = (ALCInOut*) InOut::inout;

  loopAlcMain:

  //---------------------------------------------------------------------------

  if (command == 3) { // generate a TBox

    string fileName;
    cout << "Enter the path to the parameter file: ";
    cin >> fileName;

    AlcTBoxGenType gtype;
    if (Options::readAlcGenParams(&gtype, fileName.c_str())) {
      cout << "Enter the path for output file: ";
      cin >> fileName;
      FILE *fo;
      if (!(fo = fopen(fileName.c_str(), "w"))) {
        cout << "Error on opening output file." << endl;
        deallocChildInfo();
        exit(-1);
      };

      AlcTBoxGen::writeParams(&gtype, fo);
      AlcTBoxGen::generateTBox(&gtype, fo);
      fclose(fo);
    };
  };

  //---------------------------------------------------------------------------

  if (command == 4) { // load a TBox

    FormulaCatalogue::clear();
    FormulaCatalogue::init();
    io->clearConceptDefs();

    string fileName;
    cout << "Enter the path to the TBox file: ";
    cin >> fileName;

    FILE *fi;
    if (!(fi = fopen(fileName.c_str(), "r"))) {
      cout << "Error on opening the TBox file." << endl;
      deallocChildInfo();
      exit(-1);
    };

    io->readAcyclicTBox(fi);
    fclose(fi);

    if (io->getLastConceptName())
      cout << "The last concept defined in the TBox is: "
           << *(io->getLastConceptName()) << "\n";
  };

  //---------------------------------------------------------------------------

  if (command == 5) { // Test iteractively using the loaded TBox
    string s1, s2;
    Formula *f1, *f2;

    cout << "Name of the concept standing for global assumption (TOP for true): ";
    cin >> s1;
    if (s1.compare("TOP") == 0)
      f1 = FormulaCatalogue::top;
    else
      f1 = io->getConceptDef(s1);

    cout << "Name of the concept to be tested: ";
    cin >> s2;
    f2 = io->getConceptDef(s2);

    int seconds;
    cout << "Enter the number of seconds used for the test: ";
    cin >> seconds;

    STATISTICS_INIT();
    if (Options::simplifyFormulas && f1 != FormulaCatalogue::top) {
      STATISTICS_START(simplifyingFormulas);
      f1 = f1->simplifyRecursively();
      STATISTICS_STOP(simplifyingFormulas);
    };
    clock_t ta = _stat->simplifyingFormulas[0];
    seconds -= (int) floor(((double) ta) / CLOCKS_PER_SEC);
    char rs = checkSatisfiability(f2, f1, seconds);
    STATISTICS_INC2(totalTime, ta);

    cout << (rs == SAT ? "SAT" : (rs == UNSAT ? "UNSAT" : "UNKNOWN"))
           << ": " << ((double) _stat->totalTime[0]) / CLOCKS_PER_SEC << "s\n\n";
  };

  //---------------------------------------------------------------------------

  if (command == 6) { // Test randomly using the loaded TBox

    int numOfTests;
    cout << "Number of tests: ";
    cin >> numOfTests;

    int seconds;
    cout << "Enter the number of seconds used for each single test: ";
    cin >> seconds;

    Statistics totalStat;
    clock_t cl1 = clock();
    time_t t1, t2;
    time(&t1);

    cout << "Selected pairs denote (formula_to_be_checked, global_assumption).\n";

    string s1, s2;
    Formula *f1, *f2;

    for (int i = 1; i <= numOfTests; i++) {
      char buf[20];
      string* sym;

      sprintf(buf, "C%d", rand() % io->getNumOfConceptDefs() + 1);
      s1 = buf;
      f1 = io->getConceptDef(s1);

      sprintf(buf, "C%d", rand() % io->getNumOfConceptDefs() + 1);
      s2 = buf;
      f2 = io->getConceptDef(s2);

      STATISTICS_INIT();
      if (Options::simplifyFormulas && f1 != FormulaCatalogue::top) {
        STATISTICS_START(simplifyingFormulas);
        f1 = f1->simplifyRecursively();
        STATISTICS_STOP(simplifyingFormulas);
      };
      clock_t ta = _stat->simplifyingFormulas[0];
      int seconds2 = seconds - (int) floor(((double) ta) / CLOCKS_PER_SEC);
      char rs = checkSatisfiability(f2, f1, seconds2);
      STATISTICS_INC2(totalTime, ta);

      cout << "(" << s2 << ", " << s1 << ")\t";
      cout << (rs == SAT ? "SAT" : (rs == UNSAT ? "UNSAT" : "UNKNOWN"))
             << ": " << ((double) _stat->totalTime[0]) / CLOCKS_PER_SEC << "s\n\n";
      totalStat.add(_stat);
    }; // for

    cout << "\nStatistics:\n\n";
    cout << "SAT:     " << totalStat.numOfSatResults << "\n";
    cout << "UNSAT:   " << totalStat.numOfUnsatResults << "\n";
    cout << "UNKNOWN: " << totalStat.numOfUnknownResults << "\n";

    totalStat.totalTime += ((double) clock() - cl1) / CLOCKS_PER_SEC;
    time(&t2);
    cout << "\nTotal CPU time: " << totalStat.totalTime << "s\n";
    cout << "\nTotal time: " << t2 - t1 + 1 << "s (CPU time + sleeping time)\n\n";

    if (Options::verbose)
      display_statistics();
  };

  //---------------------------------------------------------------------------

  MSG("\nAvailable commands: ");
  MSG("  3. Generate a random TBox.");
  MSG("  4. Load a TBox.");
  MSG("  5. Test iteractively using the loaded TBox.");
  MSG("  6. Test randomly using the loaded TBox.");
  MSG("  7. Exit.");
  cout << "\nPlease choose a command: ";

  while (true) {
    cin >> command;
    if (command >= 3 && command <= 7)
      break;
    else
      cout << "Enter a natural number from 3 to 7: ";
  };

  if (command != 7)
    goto loopAlcMain;

  //---------------------------------------------------------------------------

  deallocChildInfo();
  delete io;
};

//-----------------------------------------------------------------------------

float ALC::computeWeight(Formula* f)
{
  char type = f->getType();

  if (Options::weightingMethod == WM_SUMMING) {
    if (type == TOP || type == BOT)
      return 0.5;
    else if (type == POS_LITERAL || type == NEG_LITERAL)
      return 1.0;
    else if (type == FORALL || type == EXISTS) {
      float a, b;
      a = ((FormulaModalQuant*) f)->getModalIndex()->getWeight();
      b = ((FormulaModalQuant*) f)->getQuantFormula()->getWeight();
      return a + (type == FORALL ? Options::weightOfAll : Options::weightOfSome) * b;
    }
    else {
      assert(type == AND_SET || type == OR_SET);
      float weight = 0;
      set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
      for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
        weight += (*it)->getWeight();

      return weight + 
               (fs->size() - 1) * 
               (type == AND_SET ? Options::weightOfAnd : Options::weightOfOr);
    };
  }
  else if (Options::weightingMethod == WM_DIAMOND_DEPTH) {
    if (type == EXISTS)
      return 1.0 + ((FormulaModalQuant*) f)->getQuantFormula()->getWeight();
    else if (type == FORALL) {
      double k = ((FormulaModalQuant*) f)->getQuantFormula()->getWeight();
      if (k > 0) return k + 1;
      else return 0;
    }
    else if (type == AND_SET || type == OR_SET) {
      float maxWeight = 0;
      float sumOfWeights = 0;
      set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
      for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++) {
        sumOfWeights += (*it)->getWeight();
        if (maxWeight < (*it)->getWeight())
          maxWeight = (*it)->getWeight();
      };
      return Options::weightOfMaxElem * maxWeight + 
             (1.0 - Options::weightOfMaxElem) * sumOfWeights / fs->size();
    }
    else return 0;
  }
  else { // Options::weightingMethod == WM_NO_OF_WORLDS
    if (type == EXISTS)
      return 1.0 + ((FormulaModalQuant*) f)->getQuantFormula()->getWeight();
    else if (type == FORALL)
      return Options::expansionFactorOfNumOfWorlds
		* ((FormulaModalQuant*) f)->getQuantFormula()->getWeight();
    else if (type == OR_SET) {
      float maxWeight = 0;
      float sumOfWeights = 0;
      set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
      for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++) {
        sumOfWeights += (*it)->getWeight();
        if (maxWeight < (*it)->getWeight())
          maxWeight = (*it)->getWeight();
      };
      return Options::weightOfMaxElem * maxWeight + 
             (1.0 - Options::weightOfMaxElem) * sumOfWeights / fs->size();
    }
    else if (type == AND_SET) {
      float weight = 0;
      set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
      for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
        weight += (*it)->getWeight();
      return weight;
    }
    else return 0;
  };
};
