/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include <stack>
#include <map>
#include <math.h>
#include "options.h"
#include "statistics.h"
#include "cat.h"
#include "formula.h"
#include "ALC.h"

//------------------------------------------------------------------------------

map<string,string*> SymbolCatalogue::symbols;

string* SymbolCatalogue::find(string &s)
{
  map<string,string*>::iterator it;

  if((it = symbols.find(s)) == symbols.end()) {
    string* s2 = new string(s);
    symbols[*s2] = s2;
    return s2;
  } 
  else 
    return it->second;
};

//------------------------------------------------------------------------------
// hashtable-support was implemented by Lukasz Czajka

#ifdef HASHTABLE_SUPPORT

FormulaCatalogue::hashtab_t FormulaCatalogue::formulas;
Formula* FormulaCatalogue::top;
Formula* FormulaCatalogue::bot;

void FormulaCatalogue::init()
{
  assert(formulas.empty());
  //formulas.reserve();
  formulas.set_max_fill_ratio(Options::formulaCatalogueMaxFillRatio);
  FormulaCatalogue::top = new FormulaTop();
  FormulaCatalogue::bot = new FormulaBot();
  insert(top);
  insert(bot);
};

Formula* FormulaCatalogue::find(Formula *f)
{
  Formula *g;
  if ((g = formulas.insert_if_not_present(f)) != NULL) { delete f; return g; }
  else {
    assert(Formula::computeWeight);
    assert(Formula::computeWeight == ALC::computeWeight);
    f->setWeight((*Formula::computeWeight)(f));
    f->computeSize();

    f->setNegation(NULL);
    f->keepOut();
    f->setStatus(UNKNOWN);
    f->setNode(NULL);
    return f;
  };
};

void FormulaCatalogue::reduce()
{
  set<Formula*> visited;

  STATISTICS_START(compactingFormulaCatalogue);

  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end(); ++it) {
    Formula* f = *it;
    if (f->getStatus() == SAT || f->getStatus() == UNSAT || f->getNode() || f->isKeptIn())
      f->collectSubFormulas(visited);
  };

  visited.insert(bot);

/*
  for (set<Formula*>::iterator it = Graph::graph->unsatFormulas.begin();
       it != Graph::graph->unsatFormulas.end(); it++)
    (*it)->collectSubFormulas(visited);

  for (map<Formula*,Formula*>::iterator it = Graph::graph->unsatCoreMap.begin();
       it != Graph::graph->unsatCoreMap.end(); it++)
    it->second->collectSubFormulas(visited);
*/

  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end();)
    if (visited.find(*it) == visited.end()) {
      Formula* f = *it;
      assert(!f->getNode());
      it = formulas.erase(it);
      delete f;
    }
    else ++it;

  STATISTICS_STOP(compactingFormulaCatalogue);
};

void FormulaCatalogue::clear()
{
  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end(); ++it) {
    Formula* f = *it;
    if (f->getNode())
      delete f->getNode();
    delete f;
  };

  formulas.clear();
};

void FormulaCatalogue::checkAndReduce()
{
  if (formulas.size() > (unsigned) Options::minFormulasBeforeCompactingCatalogue) {
    double old_size = formulas.size();
    #ifdef VERBOSE2
    cout << "Compacting FormulaCatalogue: " << old_size;
    #endif
    reduce();
    double new_size = formulas.size();
    #ifdef VERBOSE2
    cout << " -> " << new_size << endl;
    #endif
    double gain = (old_size - new_size) / old_size;
    if (gain >= Options::expectedCatalogueCompactingGain)
      return;

    if (gain < 0.01) gain = 0.01;
    float rate = Options::expectedCatalogueCompactingGain / gain;
    if (rate > Options::catalogueExpandingRateLimit)
      rate = Options::catalogueExpandingRateLimit;

    Options::minFormulasBeforeCompactingCatalogue = (int) (old_size * rate);
  };
};

void FormulaCatalogue::preallocate(int num)
{
  formulas.reserve((int)ceil(Options::hashPreallocateFactor * sqrt(num)));
}

//------------------------------------------------------------------------------

int FormulaCatalogue::getSize2()
{
  int size = formulas.size();
  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end(); ++it)
    if ((*it)->getType() == AND_SET || (*it)->getType() == OR_SET)
      size += ((FormulaSet*) (*it))->getSetOfFormulas()->size();
    else if ((*it)->getType() == FORALL || (*it)->getType() == EXISTS)
      size++;
  return size;
};

void FormulaCatalogue::display(FILE *f)
{
  fputs("The catalogue contains: \n\n", f);
  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end(); ++it) {
    (*it)->display(f);
    fputs("\n\n", f);
  };
};

int FormulaCatalogue::getNumOfCollisions()
{
  set<uint32_t> keys;
  for (hashtab_t::iterator it = formulas.begin(); it != formulas.end(); ++it)
    keys.insert(it.get_hash());
  return formulas.size() - keys.size();
};

#else

//-----------------------------------------------------------------------------
// without hashtable support

multimap<uint32_t,Formula*> FormulaCatalogue::formulas;
Formula* FormulaCatalogue::top;
Formula* FormulaCatalogue::bot;

void FormulaCatalogue::init()
{
  assert(formulas.empty());
  FormulaCatalogue::top = new FormulaTop();
  FormulaCatalogue::bot = new FormulaBot();
  insert(top);
  insert(bot);
};

Formula* FormulaCatalogue::find0(Formula *f)
{
  multimap<uint32_t,Formula*>::iterator it;
  pair<multimap<uint32_t,Formula*>::iterator,
       multimap<uint32_t,Formula*>::iterator> ret;

  ret = formulas.equal_range(f->getHashCode());
  for (it=ret.first; it!=ret.second; ++it)
    if (it->second->equal(f))
      return it->second;

  return NULL;
};

Formula* FormulaCatalogue::find(Formula *f)
{
  Formula *g;
  if (g = find0(f)) { delete f; return g; }
  else {
    assert(Formula::computeWeight);
    assert(Formula::computeWeight == ALC::computeWeight);
    f->setWeight((*Formula::computeWeight)(f));
    f->computeSize();

    f->setNegation(NULL);
    f->keepOut();
    f->setStatus(UNKNOWN);
    f->setNode(NULL);
    return insert(f); 
  };
};

Formula* FormulaCatalogue::insert(Formula *f)
{
  formulas.insert(pair<uint32_t,Formula*>(f->getHashCode(),f));
  return f;
};

void FormulaCatalogue::reduce()
{
  set<Formula*> visited;

  STATISTICS_START(compactingFormulaCatalogue);

  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    Formula* f = it->second;
    if (f->getStatus() == SAT || f->getStatus() == UNSAT || f->getNode() || f->isKeptIn())
      f->collectSubFormulas(visited);
  };

  visited.insert(bot);

/*
  for (set<Formula*>::iterator it = Graph::graph->unsatFormulas.begin();
       it != Graph::graph->unsatFormulas.end(); it++)
    (*it)->collectSubFormulas(visited);

  for (map<Formula*,Formula*>::iterator it = Graph::graph->unsatCoreMap.begin();
       it != Graph::graph->unsatCoreMap.end(); it++)
    it->second->collectSubFormulas(visited);
*/

  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end();)
    if (visited.find(it->second) == visited.end()) {
      Formula* f = it->second;
      assert(!f->getNode());
      multimap<uint32_t,Formula*>::iterator it_ = it++;
      formulas.erase(it_);
      delete f;
    }
    else it++;

  STATISTICS_STOP(compactingFormulaCatalogue);
};

void FormulaCatalogue::clear()
{
  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    Formula* f = it->second;
    if (f->getNode())
      delete f->getNode();
    delete f;
  };

  formulas.clear();
};

void FormulaCatalogue::checkAndReduce()
{
  if (formulas.size() > Options::minFormulasBeforeCompactingCatalogue) {
    double old_size = formulas.size();
    #ifdef VERBOSE2
    cout << "Compacting FormulaCatalogue: " << old_size;
    #endif
    reduce();
    double new_size = formulas.size();
    #ifdef VERBOSE2
    cout << " -> " << new_size << endl;
    #endif

    double gain = (old_size - new_size) / old_size;
    if (gain >= Options::expectedCatalogueCompactingGain)
      return;

    if (gain < 0.01) gain = 0.01;
    float rate = Options::expectedCatalogueCompactingGain / gain;
    if (rate > Options::catalogueExpandingRateLimit)
      rate = Options::catalogueExpandingRateLimit;

    Options::minFormulasBeforeCompactingCatalogue = (int) (old_size * rate);
  };
};

//------------------------------------------------------------------------------

int FormulaCatalogue::getSize2()
{
  int size = formulas.size();
  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    if (it->second->getType() == AND_SET || it->second->getType() == OR_SET)
      size += ((FormulaSet*) it->second)->getSetOfFormulas()->size();
    else if (it->second->getType() == FORALL || it->second->getType() == EXISTS)
      size++;
  return size;
};

void FormulaCatalogue::display(FILE *f)
{
  fputs("The catalogue contains: \n\n", f);
  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    it->second->display(f);
    fputs("\n\n", f);
  };
};

int FormulaCatalogue::getNumOfCollisions()
{
  set<uint32_t> keys;
  for (multimap<uint32_t,Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    keys.insert(it->first);
  return formulas.size() - keys.size();
};

//------------------------------------------------------------------------------

#endif // HASHTABLE_SUPPORT
