/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include "options.h"
#include "formula.h"

float (*Formula::computeWeight)(Formula* f) = NULL;

//------------------------------------------------------------------------------

bool Formula::isSuperAndSetOfSome(set<Formula*> &fs, Formula* &f)
{
  assert(type != AND_SET && type != OR_SET); // not supported for AND_SET and OR_SET

  if (fs.find(this) != fs.end()) { f = this; return true; }
  return false;
};

//------------------------------------------------------------------------------

FormulaLiteral::FormulaLiteral(int _type, string* _symbol)
{
  assert(_type == POS_LITERAL || _type == NEG_LITERAL);
  type   = _type;
  symbol = _symbol;
};

void FormulaLiteral::countOccurrencesOfAtoms(
	map<Formula*,int> &pos, map<Formula*,int> &neg,
	map<Formula*,int> &box, map<Formula*,int> &dia)
{
  if (type == POS_LITERAL)
    pos[this]++;
  else
    neg[getNegation()]++;
};

bool FormulaLiteral::equal(Formula *f)
{
  if (type == f->getType())
    return symbol == ((FormulaLiteral*) f)->symbol;

  return false;
};

Formula* FormulaLiteral::getNegation()
{
  if (!negation) {
    Formula* f = FormulaCatalogue::find(new FormulaLiteral(dualType(),symbol));
    assert(!f->getNegation0());
    setNegation(f);
    f->setNegation(this);
  };
  return negation;
};

Formula* FormulaLiteral::replace(Formula* f, Formula* g)
{
  if (this == f)
    return g;

  if (type == NEG_LITERAL) 
    if (getNegation() == f)
      return g->getNegation();

  return this;
};

Formula* FormulaLiteral::replaceAt1Level(set<Formula*> &trueLits)
{
  if (trueLits.find(this) != trueLits.end())
    return FormulaCatalogue::top;
  else if (trueLits.find(getNegation()) != trueLits.end())
    return FormulaCatalogue::bot;

  return this;
};

Formula* FormulaLiteral::replaceAt1Level(Formula* atom, bool value)
{
  if (this == atom)
    return (value ? FormulaCatalogue::top : FormulaCatalogue::bot);

  if (getNegation() == atom)
    return (value ? FormulaCatalogue::bot : FormulaCatalogue::top);

  return this;
};

void FormulaLiteral::collectSubFormulas(set<Formula*> &fs)
{
  if (fs.find(this) == fs.end()) {
    fs.insert(this);
    if (negation)
      negation->collectSubFormulas(fs);
  };
};

//------------------------------------------------------------------------------

FormulaModalQuant::FormulaModalQuant(int _type, ModalIndex* _modalIndex, Formula* _qFormula)
{
  assert(_type == FORALL || _type == EXISTS);
  type = _type;
  modalIndex  = _modalIndex;
  qFormula = _qFormula;
};

bool FormulaModalQuant::checkOccurence(Formula *f)
{
  if (this == f) return true;
  return qFormula->checkOccurence(f);
};

bool FormulaModalQuant::equal(Formula *f)
{
  if (type == f->getType()) {
    FormulaModalQuant* _f = (FormulaModalQuant*) f;
    return (modalIndex == _f->modalIndex) && 
           (qFormula == _f->qFormula);
  };

  return false;
};

Formula* FormulaModalQuant::getNegation()
{
  if (!negation) {
    Formula* f = FormulaCatalogue::find(new 
                 FormulaModalQuant(dualType(), modalIndex, qFormula->getNegation()));
    assert(!f->getNegation0());
    setNegation(f);
    f->setNegation(this);
  };
  return negation;
};

Formula* FormulaModalQuant::replace(Formula* f, Formula* g)
{
  if (this == f) return g;

  Formula* qFormula2 = qFormula->replace(f, g);

  if (qFormula2 == qFormula) return this;

  FormulaModalQuant h(type, modalIndex, qFormula2);
  return h.normalize();
};

Formula* FormulaModalQuant::replaceAt1Level(set<Formula*> &trueLits)
{
  if (trueLits.find(this) != trueLits.end())
    return FormulaCatalogue::top;
  else if (trueLits.find(getNegation()) != trueLits.end())
    return FormulaCatalogue::bot;

  return this;
};

Formula* FormulaModalQuant::replaceAt1Level(Formula* atom, bool value)
{
  if (this == atom)
    return (value ? FormulaCatalogue::top : FormulaCatalogue::bot);

  if (getNegation() == atom)
    return (value ? FormulaCatalogue::bot : FormulaCatalogue::top);

  return this;
};

Formula* FormulaModalQuant::normalize()
{
  if (type == FORALL && qFormula == FormulaCatalogue::top)
    return FormulaCatalogue::top;

  if (type == EXISTS && qFormula == FormulaCatalogue::bot)
    return FormulaCatalogue::bot;

  if ((type == FORALL && qFormula->getType() == AND_SET) ||
      (type == EXISTS && qFormula->getType() ==  OR_SET)) {
    FormulaSet* f = new FormulaSet(qFormula->getType());
    set<Formula*> *fs = ((FormulaSet*) qFormula)->getSetOfFormulas();

    for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
      f->insert(FormulaCatalogue::find(new FormulaModalQuant(type, modalIndex, *it)));

    return FormulaCatalogue::find(f);
  }; // if

  return FormulaCatalogue::find(new FormulaModalQuant(*this));
};

void FormulaModalQuant::display(FILE *f)
{
  fputs(type == FORALL ? "[" : "<", f);
  modalIndex->display(f);
  fputs(type == FORALL ? "]." : ">.", f);
  qFormula->display(f);
};

void FormulaModalQuant::display()
{
  cout << (type == FORALL ? "[" : "<");
  modalIndex->display();
  cout << (type == FORALL ? "]." : ">.");
  qFormula->display();
};

void FormulaModalQuant::countOccurrencesOfAtoms(
	map<Formula*,int> &pos, map<Formula*,int> &neg,
	map<Formula*,int> &box, map<Formula*,int> &dia)
{
  if (type == FORALL)
    box[this]++;
  else
    dia[this]++;
};

void FormulaModalQuant::collectSubFormulas(set<Formula*> &fs)
{
  if (fs.find(this) == fs.end()) {
    qFormula->collectSubFormulas(fs);
    fs.insert(this);
    if (negation)
      negation->collectSubFormulas(fs);
  };
};

//------------------------------------------------------------------------------

FormulaSet::FormulaSet(int _type, Formula *f, Formula *g)
{
  type = _type;
  add(f);
  add(g);
};

bool FormulaSet::checkOccurence(Formula *f)
{
  if (this == f)
    return true;

  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    if ((*it)->checkOccurence(f))
      return true;

  return false;
};

void FormulaSet::countOccurrencesOfAtoms(
	map<Formula*,int> &pos, map<Formula*,int> &neg, 
	map<Formula*,int> &box, map<Formula*,int> &dia)
{
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    (*it)->countOccurrencesOfAtoms(pos, neg, box, dia);
};

bool FormulaSet::equal(Formula *f)
{
  if (type == f->getType())
    return formulas == ((FormulaSet*) f)->formulas;

  return false;
};

bool FormulaSet::isSuperSetOf(int _type /* AND_SET or OR_SET*/, Formula *f)
{
  if (this == f) return true;

  if (type != _type) return false;

  if (f->getType() != type)
    return (formulas.find(f) != formulas.end());

  set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();

  if (fs->size() >= formulas.size())
    return false;

  for (set<Formula*>::iterator it = fs->begin(); it != fs->end(); it++)
    if (formulas.find(*it) == formulas.end())
      return false;

  return true;
};

bool FormulaSet::isSuperAndSetOfSome(set<Formula*> &fs, Formula* &f)
{
  if (fs.find(this) != fs.end()) { f = this; return true; };

  if (type == OR_SET) return false;
  else {
    for (set<Formula*>::iterator it = fs.begin(); it != fs.end(); it++)
      if (isSuperSetOf(AND_SET, *it)) {
        f = *it;
        return true;
      };
    return false;
  };
};

bool FormulaSet::isSmallSet4CheckingSubsumption(int _type /* AND_SET or OR_SET*/)
{
  assert(_type == AND_SET ||  _type == OR_SET);
  if (type != _type) return true;
  return formulas.size() <= Options::maxSetSize4CheckingSubsumption;
};

bool FormulaSet::bottom(Formula* &cause)
{
  if (type == OR_SET) return false;

  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    Formula* neg = (*it)->getNegation();
    if (isSuperSetOf(AND_SET, neg)) {
      FormulaSet g(AND_SET, *it, neg);
      cause = g.normalize();
      return true;
    };
  };

  return false;
};

bool FormulaSet::bottom0()
{
  if (type == OR_SET) return false;

  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) {
    Formula* neg = (*it)->getNegation();
    if (isSuperSetOf(AND_SET, neg)) return true;
  };

  return false;
};

Formula* FormulaSet::getNegation()
{
  if (!negation) {
    FormulaSet* f = new FormulaSet(dualType());

    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      f->insert((*it)->getNegation());

    Formula* g = FormulaCatalogue::find(f);

    assert(!g->getNegation0());
    setNegation(g);
    g->setNegation(this);
  };
  return negation;
};

Formula* FormulaSet::replace(Formula* f, Formula* g)
{
  if (this == f)
    return g;

  FormulaSet h(type);
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    h.add((*it)->replace(f,g));

  return h.normalize();
};

#define FormulaSet_replaceAt1Level(_x_) { \
  FormulaSet* h = new FormulaSet(type); \
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++) { \
    Formula* g = _x_; \
    if (g == FormulaCatalogue::top) { \
      if (type == OR_SET) \
        { delete h; return FormulaCatalogue::top; }; \
    } \
    else if (g == FormulaCatalogue::bot) { \
      if (type == AND_SET) \
        { delete h; return FormulaCatalogue::bot; }; \
    } \
    else \
      h->add(g); \
  }; \
  if (h->getSetOfFormulas()->size() == 0) { \
    if (type == AND_SET) \
      { delete h; return FormulaCatalogue::top; } \
    else \
      { delete h; return FormulaCatalogue::bot; }; \
  }; \
  Formula* h2; \
  h->findSimpleForm(h2); \
  if (h2) { delete h; return h2; } \
  else return FormulaCatalogue::find(h); \
}

Formula* FormulaSet::replaceAt1Level(set<Formula*> &trueLits) 
{
  FormulaSet_replaceAt1Level((*it)->replaceAt1Level(trueLits));
};

Formula* FormulaSet::replaceAt1Level(Formula* atom, bool value)
{
  FormulaSet_replaceAt1Level((*it)->replaceAt1Level(atom, value));
};

void FormulaSet::add(Formula* f, bool checkSubsumption)
{
  //if (formulas.find(f) != formulas.end()) return; // this row can be deleted

  if (f->getType() != type) {
    if (!Options::checkSubsumption4NormalForm && !checkSubsumption)
      formulas.insert(f);
    else {
      for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
        if ((*it)->isSmallSet4CheckingSubsumption(dualType()))
          if (f->isSuperSetOf(dualType(), *it))
            return;

      if (f->isSmallSet4CheckingSubsumption(dualType()))
        for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); ) {
          if ((*it)->isSuperSetOf(dualType(), f)) {
            set<Formula*>::iterator it_ = it++;
            formulas.erase(it_);
            continue;
          };
          it++;
        };
      formulas.insert(f);
    }; // else
  }
  else {
    set<Formula*> *fs = ((FormulaSet*) f)->getSetOfFormulas();
    if (!Options::checkSubsumption4NormalForm && !checkSubsumption)
      formulas.insert(fs->begin(), fs->end());
    else 
      for (set<Formula*>::iterator it2 = fs->begin(); it2 != fs->end(); it2++) {
        bool bToAdd = true;
        for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
          if ((*it)->isSmallSet4CheckingSubsumption(dualType()))
            if ((*it2)->isSuperSetOf(dualType(), *it))
              { bToAdd = false; break; };
        if (bToAdd) {
          if ((*it2)->isSmallSet4CheckingSubsumption(dualType()))
            for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); ) {
              if ((*it)->isSuperSetOf(dualType(), *it2)) {
                set<Formula*>::iterator it_ = it++;
                formulas.erase(it_);
                continue;
              };
              it++;
            }; // for
          formulas.insert(*it2);
        }; // if
      }; // for
  }; // else
};

Formula* FormulaSet::normalize()
{
  assert(formulas.size() > 0);

  if (formulas.size() == 1)
    return *(formulas.begin());

  if (type == AND_SET) {
    if (formulas.find(FormulaCatalogue::bot) != formulas.end())
      return FormulaCatalogue::bot;

    set<Formula*>::iterator it = formulas.find(FormulaCatalogue::top);
    if (it != formulas.end()) {
      formulas.erase(it);
      if (formulas.size() == 1)
        return *(formulas.begin());
    };
  }
  else {
    if (formulas.find(FormulaCatalogue::top) != formulas.end())
      return FormulaCatalogue::top;

    set<Formula*>::iterator it = formulas.find(FormulaCatalogue::bot);
    if (it != formulas.end()) {
      formulas.erase(it);
      if (formulas.size() == 1)
        return *(formulas.begin());
    };
  };

  FormulaSet* f = new FormulaSet(type);
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    f->add(*it);

  return FormulaCatalogue::find(f);
};

void FormulaSet::findSimpleForm(Formula* &f)
{
  assert(formulas.size() > 0);
  f = NULL;

  if (formulas.size() == 1)
    { f = *(formulas.begin()); return; };

  if (type == AND_SET) {
    if (formulas.find(FormulaCatalogue::bot) != formulas.end())
      { f = FormulaCatalogue::bot; return; };
    if (formulas.find(FormulaCatalogue::top) != formulas.end()) {
      formulas.erase(FormulaCatalogue::top);
      if (formulas.size() == 1)
        { f = *(formulas.begin()); return; };
    };
  }
  else {
    if (formulas.find(FormulaCatalogue::top) != formulas.end())
      { f = FormulaCatalogue::top; return; };
    if (formulas.find(FormulaCatalogue::bot) != formulas.end()) {
      formulas.erase(FormulaCatalogue::bot);
      if (formulas.size() == 1)
        { f = *(formulas.begin()); return; }
    };
  };
};

void FormulaSet::display(FILE *f)
{
  fputc('(', f);
  set<Formula*>::iterator it = formulas.begin();
  (*(it++))->display(f);
  while (it != formulas.end()) {
    fputs(type == AND_SET ? " & " : " | ", f);
    (*(it++))->display(f);
  };
  fputc(')', f);
};

void FormulaSet::display()
{
  cout << '(';
  set<Formula*>::iterator it = formulas.begin();
  (*(it++))->display();
  while (it != formulas.end()) {
    cout << (type == AND_SET ? " & " : " | ");
    (*(it++))->display();
  };
  cout << ')';
};

void FormulaSet::collectSubFormulas(set<Formula*> &fs)
{
  if (fs.find(this) == fs.end()) {
    for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
      (*it)->collectSubFormulas(fs);
    fs.insert(this);
    if (negation)
      negation->collectSubFormulas(fs);
  };
};

void FormulaSet::computeSize() 
{ 
  size = formulas.size() - 1;
  for (set<Formula*>::iterator it = formulas.begin(); it != formulas.end(); it++)
    size += (*it)->getSize();
};

//------------------------------------------------------------------------------

FormulaDEF::FormulaDEF(Formula* _lhs, Formula* _rhs)
{
  type = DEF;
  lhs = _lhs;
  rhs = _rhs;
};

bool FormulaDEF::equal(Formula *f)
{
  if (type == f->getType()) {
    return lhs->equal(((FormulaDEF*) f)->get_lhs()) && 
           rhs->equal(((FormulaDEF*) f)->get_rhs());
  };

  return false;
};

Formula* FormulaDEF::normalize()
{
  if (lhs == rhs)
    return FormulaCatalogue::top;

  if (lhs->getType() == POS_LITERAL) {
    if (lhs->getNegation() == rhs)
      return FormulaCatalogue::bot;
    else
      return FormulaCatalogue::find(new FormulaDEF(*this));
  };

  if (lhs->getType() == NEG_LITERAL) {
    if (lhs->getNegation() == rhs)
      return FormulaCatalogue::bot;
    else
      return FormulaCatalogue::find(new FormulaDEF(lhs->getNegation(), rhs->getNegation()));
  };

  if (rhs->getType() == NEG_LITERAL)
    return FormulaCatalogue::find(new FormulaDEF(rhs->getNegation(), lhs->getNegation()));

  if (rhs->getType() == POS_LITERAL)
    return FormulaCatalogue::find(new FormulaDEF(rhs, lhs));

  FormulaSet f1(OR_SET, lhs->getNegation(), rhs);
  Formula* g1 = f1.normalize();

  FormulaSet f2(OR_SET, rhs->getNegation(), lhs);
  Formula* g2 = f2.normalize();

  FormulaSet g3(AND_SET, g1, g2);
  return g3.normalize();
};

Formula* FormulaDEF::replace(Formula* f, Formula* g)
{
  Formula* lhs2 = lhs->replace(f,g);
  Formula* rhs2 = rhs->replace(f,g);
  FormulaDEF h(lhs, rhs2);
  return h.normalize();
};

void FormulaDEF::collectSubFormulas(set<Formula*> &fs)
{
  if (fs.find(this) == fs.end()) {
    lhs->collectSubFormulas(fs);
    rhs->collectSubFormulas(fs);
    fs.insert(this);
    if (negation)
      negation->collectSubFormulas(fs);
  };
};

//------------------------------------------------------------------------------
