/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _FORMULA_H
#define _FORMULA_H

#include <map>
#include <set>
#include <iostream>
#include <string>
#include <stdio.h>

#include "first.h"
#include "tools.h"
#include "cat.h"
#include "mindex.h"

// formula types:
#define TOP		1
#define BOT		2
#define POS_LITERAL	3
#define NEG_LITERAL	4
#define AND_SET		5
#define OR_SET		6
#define FORALL		7
#define EXISTS		8
// for ALC with TBox:
#define DEF 		20

#include "formula0.h"

class FormulaTopBot : public Formula {
  public:
    char dualType() { return (type == TOP ? BOT : TOP); };
    void computeSize() { size = 1; };
    uint32_t getHashCode() { return hashword(type); };
    bool equal(Formula *f) { return type == f->getType(); };
    bool checkOccurence(Formula *f) { return false; }; 
    void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia) {};
    void collectSubFormulas(set<Formula*> &fs) { fs.insert(this); };

    Formula* replace(Formula* f, Formula* g) { return this; }; 
    Formula* replaceAt1Level(set<Formula*> &trueLits)    { return this; };
    Formula* replaceAt1Level(Formula* atom, bool value)  { return this; };
    Formula* normalize() { return this; };

    Formula* simplify(set<Formula*> &local, set<Formula*> &unsats) { return this; };
    Formula* sat_simplify(set<Formula*> &unsats) { return this; };
};

class FormulaTop : public FormulaTopBot {
  public:
    FormulaTop() { type = TOP; status = SAT; };
    void display(FILE *f) { fprintf(f, "TOP"); };
    void display() { cout << "TOP"; };
    Formula* getNegation() { return FormulaCatalogue::bot; };

    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true)
	{ return true; };
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true)
	{ return true; };
};

class FormulaBot : public FormulaTopBot {
  public:
    FormulaBot() { type = BOT; status = UNSAT; };
    void display(FILE *f) { fprintf(f, "BOT"); };
    void display() { cout << "BOT"; };
    Formula* getNegation() { return FormulaCatalogue::top; };
    bool bottom0() { return true; };
    bool bottom(Formula* &cause) { cause = this; return true; };

    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true)
	{ return false; };
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true)
	{ return false; };
};

//------------------------------------------------------------------------------

class FormulaLiteral : public Formula {
  protected:
    string *symbol;

  public:
    FormulaLiteral(int _type, string* _symbol);

    string* getSymbol() { return symbol; };
    char dualType() { return (type == POS_LITERAL ? NEG_LITERAL : POS_LITERAL); };
    void computeSize() { size = (type == POS_LITERAL ? 1 : 2); };
    uint32_t getHashCode() { return hashword(type, (uint32_t) symbol); };
    void display(FILE *f)
	{ fprintf(f, "%s%s", (type == NEG_LITERAL ? "~" : ""), symbol->c_str()); };
    void display()
	{ cout << (type == NEG_LITERAL ? "~" : "") << *symbol; };

    bool equal(Formula *f);
    bool checkOccurence(Formula *f) { return equal(f); };
    void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia);
    void collectSubFormulas(set<Formula*> &fs);

    Formula* getNegation();
    Formula* replace(Formula* f, Formula* g); 
    Formula* replaceAt1Level(set<Formula*> &trueLits);
    Formula* replaceAt1Level(Formula* atom, bool value);
    Formula* normalize() { return FormulaCatalogue::find(new FormulaLiteral(*this)); };

    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true);
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true);
    Formula* simplify(set<Formula*> &local, set<Formula*> &unsats);
    Formula* sat_simplify(set<Formula*> &unsats) { return FormulaCatalogue::top; };
};

//------------------------------------------------------------------------------

class FormulaModalQuant : public Formula {
  protected:
    ModalIndex* modalIndex;
    Formula* qFormula;

  public:
    FormulaModalQuant(int _type, ModalIndex* _modalIndex, Formula* _qFormula);

    Formula* getQuantFormula() { return qFormula; };
    ModalIndex* getModalIndex() { return modalIndex; };

    char dualType() { return (type == FORALL ? EXISTS : FORALL); };
    void computeSize() { size = 1 + modalIndex->getSize() + qFormula->getSize(); };
    uint32_t getHashCode() { return hashword(type, (uint32_t) modalIndex, (uint32_t) qFormula); };
    void display(FILE *f);
    void display();

    bool equal(Formula *f);
    bool checkOccurence(Formula *f);
    void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia);
    void collectSubFormulas(set<Formula*> &fs);

    Formula* getNegation();
    Formula* replace(Formula* f, Formula* g); 
    Formula* replaceAt1Level(set<Formula*> &trueLits);
    Formula* replaceAt1Level(Formula* atom, bool value);
    Formula* normalize();

    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true);
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true);
    Formula* simplify(set<Formula*> &local, set<Formula*> &unsats);
    Formula* sat_simplify(set<Formula*> &unsats); 
};

//------------------------------------------------------------------------------

class FormulaSet : public Formula {
  protected:
    set<Formula*> formulas;

  public:
    FormulaSet(int _type) { type = _type; };
    FormulaSet(int _type, Formula *f, Formula *g); 

    set<Formula*>* getSetOfFormulas() { return &formulas; };

    // For the 3 following methods, "this" is not in FormulaCatalogue.
    void erase(Formula* f) { formulas.erase(f); }; 
    void erase(set<Formula*>::iterator it) { formulas.erase(it); }; 
    void insert(Formula* f) { formulas.insert(f); };
    void add(Formula* f, bool checkSubsumption = false);

    char dualType() { return (type == AND_SET ? OR_SET : AND_SET); };
    void computeSize();
    uint32_t getHashCode() { return hashword(type, formulas); };
    void display(FILE *f);
    void display();

    bool equal(Formula *f);
    bool checkOccurence(Formula *f); 
    void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia);
    void collectSubFormulas(set<Formula*> &fs);

    Formula* getNegation();
    Formula* replace(Formula* f, Formula* g); 
    Formula* replaceAt1Level(set<Formula*> &trueLits);
    Formula* replaceAt1Level(Formula* atom, bool value); 
    bool isSuperSetOf(int _type /* AND_SET or OR_SET*/, Formula *f);
    bool isSuperAndSetOfSome(set<Formula*> &fs, Formula* &f);
    bool isSmallSet4CheckingSubsumption(int _type /* AND_SET or OR_SET*/);
    bool bottom0();
    bool bottom(Formula* &cause);
    Formula* normalize();
    void findSimpleForm(Formula* &f); // for a special use in ALCNode::_applyRule

    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true);
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true);
    Formula* simplify(set<Formula*> &local, set<Formula*> &unsats);
    Formula* sat_simplify(set<Formula*> &unsats); 
};

//------------------------------------------------------------------------------

class FormulaDEF : public Formula {
  protected:
    Formula* lhs; // "expected" form: lhs->type == POS_LITERAL
    Formula* rhs;

  public:
    FormulaDEF(Formula* _lhs, Formula* _rhs);

    Formula* get_lhs() { return lhs; };
    Formula* get_rhs() { return rhs; };

    void computeSize() { size = 1 + lhs->getSize() + rhs->getSize(); };
    uint32_t getHashCode() { return hashword(type, (uint32_t) lhs, (uint32_t) rhs); };
    void display(FILE *f) { lhs->display(f); fputs(" = ", f); rhs->display(f); };
    void display() { lhs->display(); cout << " = "; rhs->display(); };

    bool equal(Formula *f);
    void collectSubFormulas(set<Formula*> &fs);

    Formula* normalize();
    Formula* replace(Formula* f, Formula* g); 

    // unsupported methods:

    char dualType() { UNSUPPORTED(); };
    bool checkOccurence(Formula *f) { UNSUPPORTED(); };
    void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia) { UNSUPPORTED(); };
    Formula* getNegation() { UNSUPPORTED(); };
    Formula* replaceAt1Level(set<Formula*> &trueLits)   { UNSUPPORTED(); };
    Formula* replaceAt1Level(Formula* atom, bool value) { UNSUPPORTED(); };
    bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true) { UNSUPPORTED(); };
    bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true)
	{ UNSUPPORTED(); };
    Formula* simplify(set<Formula*> &local, set<Formula*> &unsats) { UNSUPPORTED(); };
    Formula* sat_simplify(set<Formula*> &unsats) { UNSUPPORTED(); }; 
};

//------------------------------------------------------------------------------

class FormulaCompare {
  public:
    bool operator()(const Formula* f, const Formula* g) const {
      if (Options::randomDFS)
        return (unsigned int) f < (unsigned int) g;

      if (Options::usingSize4SelectingDfsBranch) {
        if (f->size < g->size) return true;
        else if (f->size > g->size) return false;
        else return true;
      }
      else {
        if (f->weight < g->weight) return true;
        else if (f->weight > g->weight) return false;
        else return f->size > g->size;
      };
    };
};

#endif
