/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _FORMULA0_H
#define _FORMULA0_H

#include <map>
#include <set>
#include <iostream>

#include "first.h"

/******************************************************************************
- Formulas in FormulaCatalogue are in normalized form.
- When a formula is normalized, it is included in FormulaCatalogue.
- All subformulas are in normalized form and present in FormulaCatalogue.
- If not stated otherwise, returned formulas and formulas that are used for
  calling their methods are in normalized form and present in FormulaCatalogue.
- Formulas created by calling a consructor are not normalized yet.
- A formula is not required to be in FormulaCatalogue when calling
	- getHashCode()
	- normalize()
- A FormulaSet is not required to be in FormulaCatalogue when calling
	- erase(Formula* f)
	- insert(Formula* f)
	- add(Formula* f)
- A formula for the first parameter of Node::_apply is not required to be 
  in normalized form and present in FormulaCatalogue.
******************************************************************************/

class Formula {
  friend class FormulaCompare;

  protected:
    char type;

    char status;
    bool keptIn; // keep permantly in FormulaCatalogue
    int size;
    float weight;
    Formula* negation;
    Node *node;

  public:
    Formula() { 
	status = UNKNOWN; 
	keptIn = false; 
	size = 1; 
	weight = 0; 
	negation = NULL; 
	node = NULL; 
    };

    char getType() { return type; };
    char getStatus() { return status; };
    bool isKeptIn() { return keptIn; };
    int  getSize() { return size; };
    float getWeight() { return weight; };
    Formula* getNegation0() { return negation; }; 
    Node* getNode() { return node; };

    void setStatus(char _status) { status = _status; };
    void keepIn() { keptIn = true; };
    void keepOut() { keptIn = false; };
    void setNegation(Formula* f) { negation = f; };
    void setWeight(float _weight) { weight = _weight; };
    void setNode(Node* w) { node = w; };

    virtual char dualType() = 0;
    virtual void computeSize() = 0;
    virtual uint32_t getHashCode() = 0;
    virtual void display(FILE *f) = 0; 
    virtual void display() = 0; 

    virtual bool equal(Formula* f) = 0;
    virtual bool isSuperSetOf(int _type /* AND_SET or OR_SET*/, Formula *f) { return f == this; };
    virtual bool isSuperAndSetOfSome(set<Formula*> &fs, Formula* &f);
    virtual bool isSmallSet4CheckingSubsumption(int _type /* AND_SET or OR_SET*/) 
	{ return true; };
    virtual bool bottom0() { return false; };
    virtual bool bottom(Formula* &cause) { return false; };
    virtual bool checkOccurence(Formula *f) = 0; // check if f occurs in "this";
    virtual void countOccurrencesOfAtoms
		(map<Formula*,int> &pos, map<Formula*,int> &neg,
		 map<Formula*,int> &box, map<Formula*,int> &dia) = 0;
    virtual void collectSubFormulas(set<Formula*> &fs) = 0;

    virtual Formula* getNegation() = 0; 
    virtual Formula* replace(Formula* f, Formula* g) = 0; // replace f by g "everywhere"; 
    virtual Formula* replaceAt1Level(set<Formula*> &trueLits) = 0; 
    virtual Formula* replaceAt1Level(Formula* atom, bool value) = 0;
    virtual Formula* normalize() = 0; // "this" need not to be in FormulaCatalogue

    static float (*computeWeight)(Formula* f);

    virtual bool isSubsumedBy(Formula* f, set<Formula*> &unsats, bool flag = true) = 0;
    virtual bool isSubsumedBy(set<Formula*> &local, set<Formula*> &unsats, bool flag = true) = 0; 
	// check whether "this" is subsumed by ("f" or "local" assumptions) and
	// the global assumptions that are negations of "unsats";
	// if flag is set off, omit checking whether the negation of "this" belongs to "unsats";

    virtual Formula* simplify(set<Formula*> &local, set<Formula*> &unsats) = 0;
    Formula* simplifyRecursively(set<Formula*> &unsats);
    Formula* simplifyRecursively();
    virtual Formula* sat_simplify(set<Formula*> &unsats) = 0; 
};

//------------------------------------------------------------------------------

class FormulaHasher{
  public:
    uint32_t operator()(const Formula* f) const 
	{ return ((Formula*) f)->getHashCode(); }
};

class FormulaEqCmp{
  public:
    bool operator()(const Formula* f1, const Formula* f2) const 
	{ return ((Formula*) f1)->equal((Formula*)f2); }
};

#endif

