/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

using namespace std;

#include <math.h>
#include <map>
#include <set>
#include <queue>
#include <list>

#include "tools.h"
#include "statistics.h"
#include "cat.h"
#include "graph.h"
#include "xqueue.h"

Graph* Graph::graph = NULL;
int Graph::nextNodeID;

Graph::~Graph()
{
  delete xqueue; 
};

void Graph::init(Formula* _checkedFormula, Formula* _generalAssumptions)
{
  Graph::nextNodeID = 1;

  satProbability    = Options::satProbability;
  unsatProbability  = 1 - satProbability;
  hNumOfSatNodes    = INIT_VAL_FOR_COMP_SAT_PROB;
  hNumOfUnsatNodes  = INIT_VAL_FOR_COMP_SAT_PROB;

  statNumOfExpansions = 0;
  statNumOfEdges = 0;
  statNumOfDeletedNodes = 0;
  statNumOfDeletedEdges = 0;
  statNumOfDeletedBackEdges = 0;

  _statNumOfExpansions = 0;
  _statNumOfEdges = 0;
  _statNumOfDeletedNodes = 0;
  _statNumOfDeletedEdges = 0;
  _statNumOfDeletedBackEdges = 0;
  _hNumOfSatNodes   = INIT_VAL_FOR_COMP_SAT_PROB;
  _hNumOfUnsatNodes = INIT_VAL_FOR_COMP_SAT_PROB;

  _statNumOfExpansions2 = 0;

  unsatFormulas.insert(FormulaCatalogue::bot);

  generalAssumptions = _generalAssumptions;
  generalAssumptions->keepIn();

  initXQueue();

  FormulaSet f(AND_SET, _checkedFormula, generalAssumptions);
  Formula* tauContent = f.normalize();
  if (tauContent == FormulaCatalogue::top) { tauStatus = SAT; return; }
  if (tauContent == FormulaCatalogue::bot) { tauStatus = UNSAT; return; }

  bool bNewNode;
  getNode(tauContent, tauStatus, tau, bNewNode, true);

  if (tauStatus == UNKNOWN) {
    if (tau->status == UNEXPANDED) {
      unexpandedNodes.insert(tau);
      xqueue->push(tau);
    };
  };
}

bool Graph::checkSatisfiability()
{
  if (tauStatus == SAT) {
    STATISTICS_SET(result, SAT);
    return true;
  }
  else if (tauStatus == UNSAT) {
    STATISTICS_SET(result, UNSAT);
    return false;
  };

  while (tau->status != SAT && tau->status != UNSAT) {

    if (xqueue->empty()) {
      if (!Options::cutoff) break;
      if (!recomputeExpandingQueue()) break;
    };

    if (Options::hsExpanding)
      if (shouldRecomputeExpandingQueue())
        if (!recomputeExpandingQueue())
          break;

    Node* v = xqueue->pop();

    if (unexpandedNodes.find(v) == unexpandedNodes.end()) continue;
    if (deletedNodes.find(v->id) != deletedNodes.end()) continue;
    if (!v->reachable) continue;
    assert(v->status == UNEXPANDED);

    #ifdef TESTING
    if (Options::hsExpanding)
      cout << "Expanding a node with priority " << v->priority << " ..." << endl;
    #endif

    statNumOfExpansions++;
    STATISTICS_INC(numOfExpansions);

    doBeforeEachExpansion();

    labels_queue labels;
    set<Formula*>::iterator it;

    STATISTICS_START(choosingRules);
    bool canChooseRule = v->chooseRule(labels);
    STATISTICS_STOP(choosingRules);

    if (!canChooseRule) {
      v->status = SAT;
      updateSatProbability(SAT);
    }
    else {
      v->kind = (v->principal ? OR_NODE : AND_NODE);

      while (!labels.empty() && v->status == UNEXPANDED) {
        Formula* f;
        char status;
        Node* w;
        bool bNewNode;

	Formula* label = labels.top(); 
	labels.pop(); 

        STATISTICS_START(applyingRules);
        f = v->applyRule(label, generalAssumptions);
        STATISTICS_STOP(applyingRules);

        if (Options::subsetChecking1) {
          STATISTICS_START(subsetChecking1);
          for (map<Node*,Formula*>::iterator it = v->edges.begin(); it != v->edges.end(); it++) {
            Node* u = it->first;
            if (u->getStatus() == UNSAT)
              if (f->isSuperSetOf(AND_SET, u->getContent())) {
                f = u->getContent();
                break;
              };
          };
          STATISTICS_STOP(subsetChecking1);
        };

        getNode(f, status, w, bNewNode);

        if (status == SAT) {
          if (v->kind == OR_NODE) { 
            v->status = SAT; 
            updateSatProbability(SAT);
            break; 
          }
          else continue;
        }
        else if (status == UNSAT) {
          if (w) {
            assert(w->status == UNSAT);
            addEdge(v, w, label, bNewNode);
            if (v->kind == AND_NODE) {
              v->status = UNSAT;
              if (Options::useUnsatCores)
                childCausingUnsat[v] = w;
              updateSatProbability(UNSAT);
              break; 
            }
            else continue;
          }
          else {
            assert(f == FormulaCatalogue::bot || !Options::useUnsatCores);
            if (v->kind == AND_NODE) {
              v->status = UNSAT;
              updateSatProbability(UNSAT);
              break; 
            }
            else continue;
          };
        }
        else { // status == UNKNOWN
          if (w == v) continue;
          addEdge(v, w, label, bNewNode);
          if (w->status == UNEXPANDED) {
            if (Options::hsExpanding)
              recomputePriority(w);
            xqueue->push(w);
          };
        }; // else
      }; // for

      if (v->status == UNEXPANDED && !v->numEssentialChildren) {
        if (v->kind == AND_NODE) {
          v->status = SAT; 
          updateSatProbability(SAT);
        }
        else {
          v->status = UNSAT;
          updateSatProbability(UNSAT);
        };
      }; // if
    }; // else

    unexpandedNodes.erase(v);
    if (v->status == SAT || v->status == UNSAT) {
      set<Node*> todo;
      todo.insert(v);
      propagateStatus(todo);
    }
    else {
      v->status = EXPANDED;
      expandedNodes.insert(v);

      if (generalAssumptions != FormulaCatalogue::top)
        if (statNumOfExpansions - _statNumOfExpansions2 >= Options::minExpansionsBDSN
            && statNumOfExpansions >= _statNumOfExpansions2 
                                      * (1 + Options::minExpansionsRateBDSN)) {
          set<Node*> satNodes;
          detectSatNodes(satNodes);
          if (!satNodes.empty()) {
            DEBUG_MSG("Detected looping SAT nodes.");
            propagateStatus(satNodes);
          }
          _statNumOfExpansions2 = statNumOfExpansions;
        };
    }; // else
  }; // main "while"

  STATISTICS_SET(result, tau->status);
  return ((tau->status == UNSAT) ? false : true);
};

#define deleting_edges_from(_x) { \
    for (map<Node*,Formula*>::iterator it = _x->edges.begin(); it != _x->edges.end(); ) { \
      Node* child = it->first; \
      child->parents.erase(_x); \
      if (child->mainParent == _x) \
        child->mainParent = NULL; \
      map<Node*,Formula*>::iterator it_ = it++; \
      _x->edges.erase(it_); \
      statNumOfEdges--; \
      statNumOfDeletedEdges++; \
      if (child->parents.empty() && child != tau && \
          child->status == UNSAT) { \
        set<int>::iterator it2; \
        if ((it2 = nodesToBeDeleted.find(child->id)) != nodesToBeDeleted.end()) { \
          nodesToBeDeleted.erase(it2); \
          assert(child->edges.empty()); \
          assert(expandedNodes.find(child) == expandedNodes.end()); \
          assert(unexpandedNodes.find(child) == unexpandedNodes.end()); \
          assert(FormulaCatalogue::find0(child->content)); \
          child->content->setNode(NULL); \
          deletedNodes.insert(child->id); \
          todo.erase(child); \
          delete child; \
          statNumOfDeletedNodes++; \
        }; \
      }; \
    }; \
    assert(_x->edges.empty()); \
    _x->numEssentialChildren = 0; \
}

void Graph::propagateStatus(set<Node*> &todo)
{
  int count = 0; // just for debugging

  while (!todo.empty()) {
    Node* x = *(todo.begin());
    todo.erase(todo.begin());
    count++;

    x->content->setStatus(x->status);

    if (Options::cutoff)
      updateReachabilityRelation(x);

    // computing unsatCore
    Formula* unsatCore = x->content;
    if (x->status == UNSAT && Options::useUnsatCores) {
      map<Formula*,Formula*>::iterator it0 = unsatCoreMap.find(x->content);
      if (it0 != unsatCoreMap.end()) {
        unsatCore = it0->second;
        assert(childCausingUnsat.find(x) == childCausingUnsat.end());
      }
      else {
        map<Node*,Node*>::iterator it = childCausingUnsat.find(x);
        if (it != childCausingUnsat.end()) {
          assert(x->kind == AND_NODE);
          STATISTICS_START(computingUnsatCores);
          unsatCore = x->computeUnsatCore(generalAssumptions, it->second);
          STATISTICS_STOP(computingUnsatCores);
          childCausingUnsat.erase(it);
        }
        else {
          assert(x->kind == OR_NODE);
          STATISTICS_START(computingUnsatCores);
          unsatCore = x->computeUnsatCore(generalAssumptions);
          STATISTICS_STOP(computingUnsatCores);
        };

        map<Formula*,Formula*>::iterator it2 = unsatCoreMap.find(unsatCore);
        if (it2 != unsatCoreMap.end()) {
          assert(it2->second != unsatCore);
          assert(unsatCore->isSuperSetOf(AND_SET, it2->second));
          unsatCore = it2->second;
        };

        updateUnsatCore(x->content, unsatCore);
      }; // else
    }; // if

    deleting_edges_from(x);

    Formula* x_old_content = x->content;

    //-------------------------------------------------------------------------

    if (x->status == UNSAT && unsatCore != x->content) {
      Node* y = unsatCore->getNode();
      assert (x != y);

      // changing content of x to unsatCore
      assert(x->content->getNode() == x);
      x->content->setNode(NULL);
      x->content = unsatCore;
      unsatCore->setNode(x);
      unsatCore->setStatus(UNSAT);

      if (y) { // merging y to x
        STATISTICS_START(mergingNodes);
        if (Options::cutoff) 
          updateReachabilityRelation(y);

        deleting_edges_from(y);

        // forwarding "edges to y" to "edges to x"
        for (set<Node*>::iterator it = y->parents.begin(); it != y->parents.end(); ) {
          Node* z = *it;
          assert(z != x && z != y);

          Formula* label = z->edges[y];
          assert(label);
          z->edges.erase(y);
          if (y->status == EXPANDED || y->status == UNEXPANDED)
            z->numEssentialChildren--;
          else if (todo.find(y) != todo.end())
            z->numEssentialChildren--;

          set<Node*>::iterator it_ = it++;
          y->parents.erase(it_);

          if (z->edges.find(x) == z->edges.end())
            z->numEssentialChildren++;
          z->edges[x] = label;

          map<Node*,Node*>::iterator it2 = childCausingUnsat.find(z);
          if (it2 != childCausingUnsat.end())
            if (it2->second == y)
              childCausingUnsat[z] = x;

          x->parents.insert(z);
        };

        if (y != tau) {
          if (y->status == EXPANDED)
            expandedNodes.erase(y);
          else if (y->status == UNEXPANDED)
            unexpandedNodes.erase(y);
          assert(y->edges.empty());
          assert(y->parents.empty());
          deletedNodes.insert(y->id);
          todo.erase(y); \
          delete y;
          statNumOfDeletedNodes++;
        };
        STATISTICS_STOP(mergingNodes);
      }; // if (y)
    }; // if

    //-------------------------------------------------------------------------

    bool doneGlobalSubsetChecking2 = false;

    // propagate status by subset checking
    if (x->status == UNSAT) {
      STATISTICS_START(updatingUnsats);
      bool updatedUnsats = false;
      if (Options::subsetChecking || Options::subsetChecking2)
        updatedUnsats = updateUnsats(x->content);
      STATISTICS_STOP(updatingUnsats);
      if (updatedUnsats)
        if (Options::subsetChecking2) {
          doneGlobalSubsetChecking2 = true;
          STATISTICS_START(subsetChecking2);
          set<Node*> *pUnknownStatusNodes;
          for (int i = 0; i < 2; i++) {
            pUnknownStatusNodes = (i ? &expandedNodes : &unexpandedNodes);
            for (set<Node*>::iterator it = pUnknownStatusNodes->begin(); 
                 it != pUnknownStatusNodes->end(); ) {
              if ((*it)->content->isSuperSetOf(AND_SET, x->content)) {
                assert((*it)->status == UNEXPANDED || (*it)->status == EXPANDED);
                (*it)->status = UNSAT;
                updateUnsatCore((*it)->content, x->content);
                todo.insert(*it);
                set<Node*>::iterator it_ = it++;
                pUnknownStatusNodes->erase(it_);
                updateSatProbability(UNSAT);
              }
              else it++;
            }; // for
          }; // for
          STATISTICS_STOP(subsetChecking2);
        }; // if
    }; // if

    //-------------------------------------------------------------------------

    if (!doneGlobalSubsetChecking2)
      if (x->status == UNSAT && Options::useUnsatCores) {
        STATISTICS_START(usingUnsatCores);
        double coringEffect = x->content->getSize() / x_old_content->getSize();
        if (coringEffect <= Options::unsatCoreThreshold4SubsetChecking) {
          for (set<Node*>::iterator it = x->parents.begin(); it != x->parents.end(); it++) {
            { // a sub-block
              Node* w = *it; // a parent of x
              if (w != x && (w->status == UNEXPANDED || w->status == EXPANDED))
                if (w->content->isSuperSetOf(AND_SET, x->content)) {
                  assert(w->content != x->content);
                  if (w->status == UNEXPANDED)
                    unexpandedNodes.erase(w);
                  else
                    expandedNodes.erase(w);
                  w->status = UNSAT;
                  updateUnsatCore(w->content, x->content);
                  todo.insert(w);
                  updateSatProbability(UNSAT);
                }; // if
            }; // end of the sub-block

            for (map<Node*,Formula*>::iterator it2 = (*it)->edges.begin();
                 it2 != (*it)->edges.end(); it2++) {
              Node* w = it2->first; // a brother of x
              if (w != x && (w->status == UNEXPANDED || w->status == EXPANDED))
                if (w->content->isSuperSetOf(AND_SET, x->content)) {
                  assert(w->content != x->content);
                  if (w->status == UNEXPANDED)
                    unexpandedNodes.erase(w);
                  else
                    expandedNodes.erase(w);
                  w->status = UNSAT;
                  updateUnsatCore(w->content, x->content);
                  todo.insert(w);
                  updateSatProbability(UNSAT);
                }; // if
            }; // for
          }; // for
        }; // if
        STATISTICS_STOP(usingUnsatCores);
      }; // if


    //-------------------------------------------------------------------------

    if (x->status == SAT && Options::localSupersetChecking) {
      STATISTICS_START(localSupersetChecking);
      for (set<Node*>::iterator it = x->parents.begin(); it != x->parents.end(); it++) {
        for (map<Node*,Formula*>::iterator it2 = (*it)->edges.begin();
             it2 != (*it)->edges.end(); it2++) {
          Node* w = it2->first; // a brother of x
          if (w != x && (w->status == UNEXPANDED || w->status == EXPANDED))
            if (x->content->isSuperSetOf(AND_SET, w->content)) {
              assert(w->content != x->content);
              if (w->status == UNEXPANDED)
                unexpandedNodes.erase(w);
              else
                expandedNodes.erase(w);
              w->status = SAT;
              todo.insert(w);
              updateSatProbability(SAT);
            }; // if
        }; // for
      }; // for
      STATISTICS_STOP(localSupersetChecking);
    }; // if

    //-------------------------------------------------------------------------

    // propagating backward to parents of the considered node
    for (set<Node*>::iterator parent = x->parents.begin(); parent != x->parents.end(); ) {
      Node* u = *parent;
      Formula* label = u->edges[x];
      assert(label);

      u->numEssentialChildren--;
      statNumOfDeletedBackEdges++;

      if (x->status == SAT) { // delete the edge (u,x)
        u->edges.erase(x);
        set<Node*>::iterator parent_ = parent++;
        x->parents.erase(parent_);
        if (x->mainParent == u)
          x->mainParent = NULL;
        statNumOfEdges--;
        statNumOfDeletedEdges++;
      }
      else parent++;

      if (u->status == EXPANDED) {
        if (u->kind == OR_NODE) {
          if (x->status == SAT) {
            u->status = SAT;
            todo.insert(u);
            updateSatProbability(SAT);
          }
          else {
            assert(u->numEssentialChildren >= 0);
            if (!u->numEssentialChildren) {
              u->status = UNSAT;
              todo.insert(u);
              updateSatProbability(UNSAT);
            };
          };
        }
        else { // u->kind == AND_NODE
          if (x->status == UNSAT) {
            u->status = UNSAT;
            if (Options::useUnsatCores)
              childCausingUnsat[u] = x;
            todo.insert(u);
            updateSatProbability(UNSAT);
          }
          else {
            if (!u->numEssentialChildren) {
              u->status = SAT;
              todo.insert(u);
              updateSatProbability(SAT);
            };
          };
        };

        if (u->status != EXPANDED)
          expandedNodes.erase(u);
      }; // if (u->status == EXPANDED)
    }; // for

    //-------------------------------------------------------------------------

    assert(x->edges.empty());
    x->numEssentialChildren = 0;
    if (x != tau) {
      if (x->status == SAT || x->parents.empty()) {
        assert(x->parents.empty());
        assert(expandedNodes.find(x) == expandedNodes.end()); \
        assert(unexpandedNodes.find(x) == unexpandedNodes.end()); \
        assert(FormulaCatalogue::find0(x->content)); \
        x->content->setNode(NULL);
        deletedNodes.insert(x->id);
        delete x;
        statNumOfDeletedNodes++;
      }
      else 
        nodesToBeDeleted.insert(x->id);
    };
  }; // while

  #ifdef TESTING
  cout << "Propagation resulted in " << count-1 << " additional determined nodes.\n";
  #endif
};

#define updateNumOfEssentialChildrenOfParents(_u) { \
        if (statusNode.find(_u) == statusNode.end()) { \
          for (set<Node*>::iterator itx = (_u)->parents.begin(); \
               itx != (_u)->parents.end(); itx++) { \
            if ((*itx)->status == EXPANDED) \
              if (numEC.find(*itx) == numEC.end()) \
                numEC[*itx] = (*itx)->numEssentialChildren - 1; \
              else \
                numEC[*itx]--; \
          }; \
        }; \
}

float Graph::propagationDegree(Node* v, NodeStatus status)
{
  queue<Node*> todo;

  map<Node*,float> statusNode;
  map<Node*,int> numEC; // number of essential children
  map<Node*,float> maxChildDegree; 

  NodeKind OrIffSat = ((status == SAT) ? OR_NODE : AND_NODE);
  double statusProbability = ((status == SAT) ? satProbability : unsatProbability);

  updateNumOfEssentialChildrenOfParents(v);
  statusNode[v] = 1;
  todo.push(v);

  while (!todo.empty()) {
    Node* x = todo.front();
    todo.pop();

    for (set<Node*>::iterator parent = x->parents.begin(); parent != x->parents.end(); parent++) {
      Node* u = *parent;
      if (u->status == EXPANDED) {
        if (u->kind == OrIffSat) {
          if (statusNode[x] - statusNode[u] >= Options::propagationDegreeGranule) {
            updateNumOfEssentialChildrenOfParents(u);
            statusNode[u] = statusNode[x];
            todo.push(u);
          };
        }
        else {
          float newDegree;

          if (maxChildDegree.find(u) == maxChildDegree.end())
            maxChildDegree[u] = statusNode[x];
          else if (maxChildDegree[u] < statusNode[x])
            maxChildDegree[u] = statusNode[x];

          newDegree = maxChildDegree[u] * exp(numEC[u]*log(statusProbability));

          if (newDegree - statusNode[u] >= Options::propagationDegreeGranule) {
            updateNumOfEssentialChildrenOfParents(u);
            statusNode[u] = newDegree;
            todo.push(u);
          }; // if
        }; // else
      }; // if
    }; // for
  }; // while

  float rs = 0;
  for (map<Node*,float>::iterator it = statusNode.begin(); it != statusNode.end(); it++) 
    rs += it->second;
  return rs - 1;
};

void Graph::detectSatNodes(set<Node*> &satNodes)
{
  STATISTICS_START(detectingSatNodes);
  set<Node*> visited = unexpandedNodes;
  list<Node*> todo(unexpandedNodes.begin(), unexpandedNodes.end());

  while (!todo.empty()) {
    Node* x = todo.front(); 
    todo.pop_front(); 

    for (set<Node*>::iterator it = x->parents.begin(); it != x->parents.end(); it++)
      if (visited.find(*it) == visited.end()) {
        todo.push_front(*it);
        visited.insert(*it);
      };
  }; // while

  assert(satNodes.empty());

  for (set<Node*>::iterator it = expandedNodes.begin(); it != expandedNodes.end(); )
    if (visited.find(*it) == visited.end()) {
      (*it)->status = SAT;
      satNodes.insert(*it);
      set<Node*>::iterator _it = it++;
      expandedNodes.erase(_it);
    }
    else it++;

  updateSatProbability(SAT, satNodes.size());
  STATISTICS_STOP(detectingSatNodes);
};

void Graph::updateReachabilityRelation(Node* w) // when w becomes unreachable
// Here: cutoff the reachability subtree
{
  queue<Node*> todo;
  w->reachable = false;
  todo.push(w);

  STATISTICS_START(updatingReachabilityRelation);

  while (!todo.empty()) {
    Node* x = todo.front();
    todo.pop();
    for (map<Node*,Formula*>::iterator it = x->edges.begin(); it != x->edges.end(); it++) {
      Node* u = it->first;
      if (u->mainParent == x && u->reachable) {
        u->reachable = false;
        todo.push(u);
      }; 
    }; // for
  }; // while

  STATISTICS_STOP(updatingReachabilityRelation);
};

#define get_new_unsat_node(w,g) { \
      if (g->getNode()) \
        w = g->getNode(); \
      else { \
        bNewNode = true; \
        w = newNode(g); \
        g->setNode(w); \
      }; \
      w->status = UNSAT; \
      g->setStatus(UNSAT); \
      status = UNSAT; \
}

void Graph::getNode(Formula* f, char &status, Node* &w, bool &bNewNode, bool forTau)
{
  bNewNode = false;
  w = f->getNode();

  if (w) {
    assert(w->content == f);
    assert(w->status != SAT);
    status = (w->status == UNSAT ? UNSAT : UNKNOWN);
    if (status == UNSAT)
      assert(f->getStatus() == UNSAT);
    return;
  };

  if (f->getStatus() == SAT) { status = SAT; return; };

  if (f->getStatus() == UNSAT) {
    if (!Options::useUnsatCores) 
      { status = UNSAT; return; }
    else {
      Formula* g = f;
      map<Formula*,Formula*>::iterator it = unsatCoreMap.find(f);
      if (it != unsatCoreMap.end())
        g = it->second;

      get_new_unsat_node(w,g);
      return;
    };
  };

  Formula* g;
  if (!Options::useUnsatCores) {
    if (f->bottom0())
      { status = UNSAT; return; };
    if (Options::subsetChecking) {
      STATISTICS_START(subsetChecking);
      bool isUnsat = f->isSuperAndSetOfSome(unsatFormulas, g);
      STATISTICS_STOP(subsetChecking);
      if (isUnsat) { status = UNSAT; return; };
    };
  }
  else {
    bool bBot = f->bottom(g);

    if (!bBot && Options::subsetChecking) {
      STATISTICS_START(subsetChecking);
      bBot = f->isSuperAndSetOfSome(unsatFormulas, g);
      STATISTICS_STOP(subsetChecking);
    }

    if (bBot) {
      get_new_unsat_node(w,g);
      return;
    };
  };

  bNewNode = true;
  w = newNode(f);
  f->setNode(w);
  unexpandedNodes.insert(w);
  status = UNKNOWN;
};

void Graph::addEdge(Node* u, Node* v, Formula* edgeLabel, bool bSkeletonEdge)
{
  if (v->status != SAT && v->status != UNSAT)
    if (u->edges.find(v) == u->edges.end())
      u->numEssentialChildren++;

  u->edges[v] = edgeLabel;
  v->parents.insert(u);
  statNumOfEdges++;
  if (bSkeletonEdge) {
    v->mainParent = u;
    v->depth = u->depth + 1;
  };
};

bool Graph::recomputeExpandingQueue() 
// when is succesful, xqueue must contain a node reachable and unexpanded
{
  set<Node*> visited;
  ExpandingQueue *todo;

  STATISTICS_START(recomputingExpandingQueue);
  VERBOSE_MSG("Rebuilding expanding queue: BEGIN");
  _statNumOfExpansions = statNumOfExpansions;
  _statNumOfEdges = statNumOfEdges;
  _statNumOfDeletedNodes = statNumOfDeletedNodes;
  _statNumOfDeletedBackEdges = statNumOfDeletedBackEdges;
  _hNumOfSatNodes = hNumOfSatNodes;
  _hNumOfUnsatNodes = hNumOfUnsatNodes;

  delete xqueue;
  deletedNodes.clear();
  initXQueue();

  if (Options::dfsExpanding)
    todo = new dfsExpandingQueue(this);
  else // BFS or HS
    todo = new bfsExpandingQueue();

  todo->push(tau);

  while (!todo->empty()) {
    Node* u = todo->pop();
    if (visited.find(u) != visited.end())
      continue;
    visited.insert(u);
    u->reachable = true; 
    for (map<Node*,Formula*>::iterator it = u->edges.begin(); it != u->edges.end(); it++) {
      Node* v = it->first;
      if (v->status != SAT && v->status != UNSAT && visited.find(v) == visited.end()) {
        v->mainParent = u;
        v->depth = u->depth + 1;
        if (v->status == UNEXPANDED) {
          if (Options::hsExpanding)
            recomputePriority(v);
          xqueue->push(v);
        };
        todo->push(v);
      }; // if
    }; // for
  }; // while

  delete todo;
  VERBOSE_MSG("Rebuilding expanding queue: END");
  STATISTICS_STOP(recomputingExpandingQueue);
  return !xqueue->empty();
};

void Graph::recomputePriority(Node* v)
{
  double prop_deg = 0;

  STATISTICS_START(computingNodePriorities);

  if (Options::priorityOfPropagationDegree) {
    prop_deg = satProbability * propagationDegree(v, SAT) +
               unsatProbability * propagationDegree(v, UNSAT);
    if (Options::propagationDegreeGranule)
      prop_deg = trunc(prop_deg / Options::propagationDegreeGranule) * 
                 Options::propagationDegreeGranule;
  };

  v->priority =
	Options::priorityOfPropagationDegree * prop_deg +
	Options::priorityOfContentWeight *
		trunc(((double) v->content->getWeight()) / tau->content->getWeight()
		      * Options::numOfGroupsOfFormulas)
		/ Options::numOfGroupsOfFormulas +
	Options::priorityOfContentSize *
		trunc(((double) v->content->getSize()) / tau->content->getSize()
		      * Options::numOfGroupsOfFormulas)
		/ Options::numOfGroupsOfFormulas +
	Options::priorityOfDFS * (v->id - Graph::nextNodeID) / Graph::nextNodeID + 
	Options::priorityOfNodeDepth * v->depth;

  STATISTICS_STOP(computingNodePriorities);
}; 

bool Graph::shouldRecomputeExpandingQueue()
{
  if (!Options::priorityOfPropagationDegree)
    return false;

  if (statNumOfExpansions - _statNumOfExpansions < Options::minExpansionsBREQ)
    return false;

  if (statNumOfExpansions < _statNumOfExpansions * (1 + Options::minExpansionsRateBREQ))
    return false;

  if (statNumOfDeletedBackEdges - _statNumOfDeletedBackEdges <
      _statNumOfEdges * Options::minDeletedBackEdgesRateBREQ)
    return false;

  return true;
};

void Graph::initXQueue()
{
  if (Options::dfsExpanding)
    xqueue = new dfsExpandingQueue(this);
  else if (Options::bfsExpanding)
    xqueue = new bfsExpandingQueue();
  else
    xqueue = new hsExpandingQueue();
};

void Graph::updateSatProbability(NodeStatus addedStatus, int count)
{
  if (!Options::weightOfDynamicSatProbability)
    return;

  if (addedStatus == SAT)
    hNumOfSatNodes += count;
  else {
    assert(addedStatus == UNSAT);
    hNumOfUnsatNodes += count;
  };

  double dynamicSatProbability = hNumOfSatNodes / (hNumOfSatNodes + hNumOfUnsatNodes);

  if (dynamicSatProbability > 0.5)
    satProbability = 0.5 + Options::weightOfDynamicSatProbability * (dynamicSatProbability - 0.5);
  else 
    satProbability = 0.5 - Options::weightOfDynamicSatProbability * (0.5 - dynamicSatProbability);

  unsatProbability = 1.0 - satProbability;
};

void Graph::updateUnsatCore(Formula* f, Formula* unsatCore)
{
  if (f == unsatCore) return;

  assert(f->isSuperSetOf(AND_SET, unsatCore));

  map<Formula*,Formula*>::iterator itu = unsatCoreMap.find(unsatCore);
  if (itu != unsatCoreMap.end())
    unsatCore = itu->second;

  map<Formula*,Formula*>::iterator itf = unsatCoreMap.find(f);
  if (itf == unsatCoreMap.end())
    unsatCoreMap[f] = unsatCore;
  else if (!unsatCore->isSuperSetOf(AND_SET, itf->second))
    unsatCoreMap[f] = unsatCore;
};

bool Graph::updateUnsats(Formula* f)
{
  int len;
  if (f->getType() == AND_SET) {
    len = ((FormulaSet*) f)->getSetOfFormulas()->size();
    if ( len > Options::maxLenLimitForKeptUnsat)
      return false;
  }
  else len = 1;

  set<Formula*>::iterator it = unsatFormulas.begin();
  set<Formula*>::iterator it0 = it;

  while (it != unsatFormulas.end()) {
    if (f->isSuperSetOf(AND_SET, *it)) 
      return false;
    if ((*it)->getType() == AND_SET)
      if (((FormulaSet*) *it)->getSetOfFormulas()->size() > len)
        break;
    it0 = it++;
  };

  while (it != unsatFormulas.end()) {
    if ((*it)->isSuperSetOf(AND_SET, f)) {
      set<Formula*>::iterator it_ = it++;
      unsatFormulas.erase(it_);
    }
    else it++;
  };

  if (it0 == unsatFormulas.end())
    unsatFormulas.insert(it0, f);
  else
    unsatFormulas.insert(++it0, f);

  return true;
};
