/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _GRAPH_H
#define _GRAPH_H

#include <set>

#include "first.h"
#include "options.h"
#include "formula.h"
#include "node.h"

#define INIT_VAL_FOR_COMP_SAT_PROB 200

class Graph {
  friend class dfsExpandingQueue;
  friend class FormulaCatalogue;
  friend class ALCNode;

  public:
    static Graph *graph;
    static int nextNodeID;

  protected:
    Formula* generalAssumptions;

    // the graph:
    Node* tau; // the root of the graph
    char tauStatus; // for a temporary use only
    set<Node*> unexpandedNodes;
    set<Node*> expandedNodes;
    set<int> deletedNodes; // for dealing with xqueue
    set<int> nodesToBeDeleted;

    ExpandingQueue* xqueue;

    set<Formula*> unsatFormulas;
    map<Node*,Node*> childCausingUnsat; // reading: childCausingUnsat[parent]
    map<Formula*,Formula*> unsatCoreMap; 

    // for heuristic search:
    double satProbability;
    double unsatProbability;
    int hNumOfSatNodes;
    int hNumOfUnsatNodes;

    // statistics for heuristic search:
    int statNumOfExpansions;
    int statNumOfEdges;
    int statNumOfDeletedNodes;
    int statNumOfDeletedEdges;
    int statNumOfDeletedBackEdges;

    // at the last time of re-computation of "xqueue":
    int _statNumOfExpansions;
    int _statNumOfEdges;
    int _statNumOfDeletedNodes;
    int _statNumOfDeletedEdges;
    int _statNumOfDeletedBackEdges;
    int _hNumOfSatNodes;
    int _hNumOfUnsatNodes;

    // at the last time of detectSatNodes:
    int _statNumOfExpansions2;

  public:
    ~Graph();

    void init(Formula* _checkedFormula, Formula* _generalAssumptions);
    bool checkSatisfiability();

  protected:
    virtual Node* newNode(Formula* f) = 0; 
    virtual void doBeforeEachExpansion() { FormulaCatalogue::checkAndReduce(); };

    void initXQueue();
    void updateUnsatCore(Formula* f, Formula* unsatCore);
    bool updateUnsats(Formula* f);
    set<Formula*>* getUnsatFormulas() { return &unsatFormulas; };
    void getNode(Formula* f, char &status, Node* &w, bool &bNewNode, bool forTau = false); 
	// status is one of UNKNOWN, SAT, UNSAT;
	// w and bNewNode are returned only when status != SAT;
    void addEdge(Node* u, Node* v, Formula* edgeLabel, bool bSkeletonEdge); 
    void propagateStatus(set<Node*> &todo);
    void detectSatNodes(set<Node*> &satNodes);
    virtual bool shouldRecomputeExpandingQueue(); // checking Options::hsExpanding is done outside
    virtual bool recomputeExpandingQueue(); 
	// when is succesful, xqueue must contain a node reachable and unexpanded

    virtual void updateSatProbability(NodeStatus addedStatus, int count = 1);
    virtual float propagationDegree(Node* v, NodeStatus status);
    virtual void recomputePriority(Node* v); // checking Options::hsExpanding is done outside
    virtual void updateReachabilityRelation(Node* w); // when w becomes unreachable
};

#endif
