/* provided by Lukasz Czajka */

#ifndef HASHTABLE_H
#define HASHTABLE_H

#include <iterator>
#include <vector>
#include <stdint.h>

//#define TESTING_HASH

// elements of this hashtable should be pointers
template <class T, class Hasher, class EqCmp>
class Hashtable{
private:
  struct Hash_node{
    uint32_t hash;
    T* elem;
    Hash_node* next;
  };

public:
  class iterator : public std::iterator<std::forward_iterator_tag, T*>{
  public:
    friend class Hashtable;

    iterator() : _pos(NULL), _index(INT_MAX) {}
    iterator(Hash_node* pos, unsigned index, Hash_node* prev_pos, Hashtable* tab) :
      _pos(pos), _index(index), _prev_pos(prev_pos), _tab(tab)
    { if (_pos == NULL) { ++_index; go_to_nearest_valid(); } }

    iterator& operator++() { inc(); return *this; }
    iterator operator++(int) { iterator itr(*this); inc(); return itr; }

    T* operator*() const { return _pos->elem; }
    T** operator->() const { return &_pos->elem; }

    bool operator==(const iterator& i2) const { return _pos == i2._pos; }
    bool operator!=(const iterator& i2) const { return _pos != i2._pos; }

    uint32_t get_hash() const { return _pos->hash; }

  private:
    Hash_node* _pos;
    unsigned _index;
    Hash_node* _prev_pos;
    Hashtable* _tab;

    void inc();
    void go_to_nearest_valid();
  };

public:
  Hashtable();
  ~Hashtable();

  Hashtable(const Hashtable& htab) {
    _node_pool = NULL;
    _node_pool_size = 0;
    copy(&htab);
  };
  Hashtable& operator=(const Hashtable& htab) { copy(&htab); return *this; }

  bool empty() const { return _size == 0; }
  unsigned size() const { return _size; }
  iterator begin() { return iterator(_v[0], 0, NULL, this); }
  iterator end() { return iterator(NULL, _v.size(), NULL, this); }

  // insert doesn't check whether elem is already in the table
  void insert(T* elem);
  // insert_if_not_present does and returns NULL if actually inserted,
  // or the element in the table equal to elem
  T* insert_if_not_present(T* elem);
  // returns NULL if not found
  T* find_elem(const T* elem);
  iterator find(const T* elem);
  unsigned count(const T* elem);
  iterator erase(const iterator& itr);
  // returns the number of erased elements
  unsigned erase(const T* elem);

  template <class Itr>
  void insert(Itr beg, Itr end)
  { // a hack, should check_rehash() only once before the loop
    for (; beg != end; ++beg)
      insert(*beg);
  }

  void reserve(unsigned s);

  void clear();

  float max_fill_ratio() const { return _max_fill_ratio; }
  void set_max_fill_ratio(float x) { _max_fill_ratio = x; }

  bool operator==(Hashtable& htab)
  { // a hack
    if (_size != htab.size())
      return false;
    for (iterator itr = begin(); itr != end(); ++itr)
    {
      if (htab.count(*itr) != count(*itr))
        return false;
    }
    return true;
  }

private:
  std::vector<Hash_node*> _v;
  Hasher _hash;
  EqCmp _equal;
  unsigned _size;
  unsigned _log_v_size;
  uint32_t _hash_mask;
  Hash_node* _node_pool;
  unsigned _node_pool_size;
  unsigned _max_node_pool_size;
  unsigned _next_rehash_size;
  float _max_fill_ratio;

  void copy(const Hashtable* htab);

  void rehash(int grow_rate);
  void check_rehash()
  {
    if (_size > _next_rehash_size)
      {
        rehash(1);
      }
  }
  void recount_params();

  Hash_node* new_node();
  void free_node(Hash_node* node);
};

//-----------------------------------------------------------------------------
// -*- coding: utf-8 -*-

#include <cassert>
#include <cmath>
#include <stdexcept>

// MAX_FILL_RATIO should be changed using Options::formulaCatalogueMaxFillRatio
#define MAX_FILL_RATIO 0.8
#define MAX_NODE_POOL_SIZE_FACTOR 0.2
#define INITIAL_LOG_SIZE 6

extern unsigned ceil_log2(unsigned);

// iterator

template<class T, class Hasher, class EqCmp>
inline void Hashtable<T,Hasher,EqCmp>::iterator::inc()
{
  _prev_pos = _pos;
  _pos = _pos->next;
  if (_pos == NULL)
    {
      ++_index;
      _prev_pos = NULL;
      go_to_nearest_valid();
    }
}

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::iterator::go_to_nearest_valid()
{
  assert (_pos == NULL);
  while (_index < _tab->_v.size() && _tab->_v[_index] == NULL)
    {
      ++_index;
    }
  if (_index < _tab->_v.size())
    _pos = _tab->_v[_index];
  else
    _pos = NULL;
}


// members of Hashtable

template<class T, class Hasher, class EqCmp>
Hashtable<T,Hasher,EqCmp>::Hashtable()
{
  _log_v_size = INITIAL_LOG_SIZE;
  _size = 0;
  _node_pool = NULL;
  _node_pool_size = 0;
  _max_fill_ratio = MAX_FILL_RATIO;
  _v.resize(1 << _log_v_size);
  recount_params();
}

template<class T, class Hasher, class EqCmp>
Hashtable<T,Hasher,EqCmp>::~Hashtable()
{
  for (unsigned i = 0; i < _v.size(); ++i)
    {
      Hash_node* next;
      for (Hash_node* node = _v[i]; node != NULL; node = next)
        {
          next = node->next;
          delete node;
        }
    }
  Hash_node* next;
  for (Hash_node* node = _node_pool; node != NULL; node = next)
    {
      next = node->next;
      delete node;
    }
}

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::insert(T* elem)
{
  check_rehash();
  Hash_node* node = new_node();
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  node->hash = hash;
  node->elem = elem;
  node->next = _v[ind];
  _v[ind] = node;
  ++_size;
  assert (find(elem) != end());
}

template<class T, class Hasher, class EqCmp>
T* Hashtable<T,Hasher,EqCmp>::insert_if_not_present(T* elem)
{
  check_rehash();
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  Hash_node* node = _v[ind];
  while (node != NULL)
    {
      if (node->hash == hash && _equal(node->elem, elem))
        return node->elem;
      node = node->next;
    }
  node = new_node();
  node->hash = hash;
  node->elem = elem;
  node->next = _v[ind];
  _v[ind] = node;
  ++_size;
  assert (find(elem) != end());
  return NULL;
}

template<class T, class Hasher, class EqCmp>
T* Hashtable<T,Hasher,EqCmp>::find_elem(const T* elem)
{
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  Hash_node* node = _v[ind];
  while (node != NULL)
    {
      if (node->hash == hash && _equal(node->elem, elem))
        return node->elem;
      node = node->next;
    }
  return NULL;
}

template<class T, class Hasher, class EqCmp>
typename Hashtable<T,Hasher,EqCmp>::iterator Hashtable<T,Hasher,EqCmp>::find(const T* elem)
{
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  Hash_node* node = _v[ind];
  Hash_node* prev = NULL;
  while (node != NULL)
    {
      if (node->hash == hash && _equal(node->elem, elem))
        return iterator(node, ind, prev, this);
      prev = node;
      node = node->next;
    }
  return end();
}

template<class T, class Hasher, class EqCmp>
unsigned Hashtable<T,Hasher,EqCmp>::count(const T* elem)
{
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  Hash_node* node = _v[ind];
  unsigned count = 0;
  while (node != NULL)
  {
    if (node->hash == hash && _equal(node->elem, elem))
      ++count;
    node = node->next;
  }
  assert (count == 0 || find(elem) != end());
  return count;
}

template<class T, class Hasher, class EqCmp>
typename Hashtable<T,Hasher,EqCmp>::iterator Hashtable<T,Hasher,EqCmp>::erase(const iterator& itr)
{
  Hash_node* pos = itr._pos;
  unsigned ind = itr._index;
  assert (ind < _v.size());
  --_size;
  if (pos == _v[ind])
    {
      _v[ind] = pos->next;
      free_node(pos);
      return iterator(_v[ind], ind, NULL, this);
    }
  else
    {
      Hash_node* prev = itr._prev_pos;
      prev->next = pos->next;
      free_node(pos);
      return iterator(prev->next, ind, prev, this);
    }
}

template<class T, class Hasher, class EqCmp>
unsigned Hashtable<T,Hasher,EqCmp>::erase(const T* elem)
{
  uint32_t hash = _hash(elem);
  unsigned ind = hash & _hash_mask;
  Hash_node* node = _v[ind];
  Hash_node* prev = NULL;
  Hash_node* next;
  bool erased_num = 0;
  while (node != NULL)
    {
      if (node->hash == hash && _equal(node->elem, elem))
        {
          ++erased_num;
          --_size;
          if (node == _v[ind])
            {
              _v[ind] = node->next;
              free_node(node);
              next = _v[ind];
            }
          else
            {
              assert (prev != NULL);
              prev->next = node->next;
              free_node(node);
              next = prev->next;
            }
        }
      else
        {
          prev = node;
          next = node->next;
        }
      node = next;
    }
  assert (find(elem) == end());
  assert (count(elem) == 0);
  return erased_num;
}

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::reserve(unsigned s)
{
  unsigned gr = ceil_log2(s) - _log_v_size;
  if (gr > 0)
    rehash(gr);
}

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::clear()
{
  for (unsigned i = 0; i < _v.size(); ++i)
    {
      Hash_node* next;
      for (Hash_node* node = _v[i]; node != NULL; node = next)
        {
          next = node->next;
          free_node(node);
        }
      _v[i] = NULL;
    }
  _log_v_size = INITIAL_LOG_SIZE;
  _size = 0;
  _v.resize(1 << _log_v_size);
  recount_params();
}

// private members of Hashtable

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::copy(const Hashtable* htab)
{
  _hash = htab->_hash;
  _equal = htab->_equal;
  _size = htab->_size;
  _log_v_size = htab->_log_v_size;
  _hash_mask = htab->_hash_mask;
  _max_node_pool_size = _size; // to avoid deleting nodes in this function
  _next_rehash_size = htab->_next_rehash_size;
  _max_fill_ratio = htab->_max_fill_ratio;
  _v.resize(htab->_v.size());
  for (unsigned i = 0; i < _v.size(); ++i)
    {
      Hash_node* node = _v[i];
      while (node != NULL)
        {
          Hash_node* next = node->next;
          free_node(node);
          node = next;
        }
      Hash_node* node2 = htab->_v[i];
      if (node2 != NULL)
        {
          _v[i] = new_node();
          node = _v[i];
          while (node2->next != NULL)
            {
              *node = *node2;
              node->next = new_node();
              node = node->next;
              node2 = node2->next;
            }
          *node = *node2;
          node->next = NULL;
        }
      else
        {
          _v[i] = NULL;
        }
    }
  _max_node_pool_size = htab->_max_node_pool_size;
  assert (*((Hashtable*) this) == *((Hashtable*) htab));
}

template<class T, class Hasher, class EqCmp>
void Hashtable<T,Hasher,EqCmp>::rehash(int grow_rate)
{
  if (grow_rate == 0)
    return;
  Hash_node* chain = NULL;
#ifdef TESTING_HASH
  cout << "rehash: " << grow_rate << endl;
  double avg_count = 0.0;
  double m2_count = 0.0;
  int max_count = 0;
  int nonempty = 0;
#endif
  for (unsigned i = 0; i < _v.size(); ++i)
    {
      Hash_node* node = _v[i];
#ifdef TESTING_HASH
      int count = 0;
#endif
      while (node != NULL)
        {
#ifdef TESTING_HASH
          ++count;
#endif
          Hash_node* next = node->next;
          node->next = chain;
          chain = node;
          node = next;
        }
      _v[i] = NULL;
#ifdef TESTING_HASH
      avg_count += count;
      m2_count += count * count;
      if (count > max_count)
        max_count = count;
      if (count != 0)
        ++nonempty;
#endif
    }
#ifdef TESTING_HASH
  avg_count /= nonempty;
  m2_count /= nonempty;
  cout << "average nonempty chain length: " << avg_count << endl;
  cout << "deviation: " << sqrt(m2_count - avg_count * avg_count) << endl;
  cout << "max chain length: " << max_count << endl;
#endif
  _log_v_size += grow_rate;
  _v.resize(1 << _log_v_size);
  recount_params();
  while (chain != NULL)
    {
      Hash_node* next = chain->next;
      unsigned ind = chain->hash & _hash_mask;
      chain->next = _v[ind];
      _v[ind] = chain;
      chain = next;
    }
#ifdef TESTING_HASH
  cout << "rehash end" << endl;
#endif
}

template<class T, class Hasher, class EqCmp>
inline void Hashtable<T,Hasher,EqCmp>::recount_params()
{
  _max_node_pool_size = (int)std::floor(MAX_NODE_POOL_SIZE_FACTOR * (double)(1 << _log_v_size));
  _next_rehash_size = (int)std::floor(_max_fill_ratio * (double)(1 << _log_v_size));
  _hash_mask = (1 << _log_v_size) - 1;
}

template<class T, class Hasher, class EqCmp>
inline typename Hashtable<T,Hasher,EqCmp>::Hash_node* Hashtable<T,Hasher,EqCmp>::new_node()
{
  if (_node_pool != NULL)
    {
      Hash_node* node = _node_pool;
      _node_pool = _node_pool->next;
      --_node_pool_size;
      return node;
    }
  else
    return new Hash_node;
}

template<class T, class Hasher, class EqCmp>
inline void Hashtable<T,Hasher,EqCmp>::free_node(Hash_node* node)
{
  if (_node_pool_size < _max_node_pool_size)
    {
      node->next = _node_pool;
      _node_pool = node;
      ++_node_pool_size;
    }
  else
    delete node;
}

#endif
