/***************************************************************************
 *   Copyright (C) 2008 by dr Linh Anh Nguyen                              *
 *   Institute of Informatics, University of Warsaw                        *
 *   nguyen@mimuw.edu.pl                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef _XQUEUE_H
#define _XQUEUE_H

#include <set>
#include <stack>
#include <queue>

#include "options.h"
#include "node.h"
#include "graph.h"

/***************************************************************************
* Dependencies:
*	- Classes: Node, Graph
*	- Members of Node: priority
*	- Members of Graph: statNumOfNodes
***************************************************************************/

class ExpandingQueue {
  public:
    virtual bool empty() = 0;
    virtual Node* pop() = 0;
    virtual void push(Node* w) = 0;
    virtual int size() = 0;
};

class dfsExpandingQueue: public ExpandingQueue { // for depth-first-search
  protected:
    stack<Node*> lista;
    Graph *graph;

    void compact() {
	set<Node*> elms;
	stack<Node*> lista2;
	while (!lista.empty()) {
	  Node* w = lista.top();
	  lista.pop();
	  if (elms.find(w) == elms.end()) {
	    lista2.push(w);
	    elms.insert(w);
	  };
	};
	while (!lista2.empty()) {
	  Node* w = lista2.top();
	  lista2.pop();
	  lista.push(w);
	};
    };

  public:
    dfsExpandingQueue(Graph* g) { graph = g; };
    bool empty() { return lista.empty(); };
    int size() { return lista.size(); };

    Node* pop() { Node* w = lista.top(); lista.pop(); return w; };
    void push(Node* w) { 
	lista.push(w); 
	if (lista.size() > Options::dfsExpandingQueueSizeThreshold)
          if (lista.size() > Options::dfsExpandingQueueSizeRatio * 
		(graph->unexpandedNodes.size() + graph->expandedNodes.size())) {
            #ifdef VERBOSE2
            cout << "Compacting Expanding Stack: " << lista.size();
            #endif
	    compact();
            #ifdef VERBOSE2
            cout << " -> " << lista.size() << endl;
            #endif
          };
    };
};

class bfsExpandingQueue: public ExpandingQueue { // for breadth-first-search
  protected:
    queue<Node*> lista;
    set<Node*> elms; // another representation of "lista"

  public:
    bfsExpandingQueue() {};
    bool empty() { return lista.empty(); };
    int size() { return lista.size(); };

    Node* pop() { 
	Node* w = lista.front(); 
	lista.pop(); 
	elms.erase(w); 
	return w; 
    };

    void push(Node* w) { 
	if (elms.find(w) == elms.end()) { 
	  lista.push(w); 
	  elms.insert(w); 
	}; 
    };
};

class NodeCompare {
  public:
    bool operator()(const Node* x, const Node* y) const { return x->priority < y->priority; };
};

class hsExpandingQueue: public ExpandingQueue { // for heuristic-search
  protected:
    priority_queue<Node*,vector<Node*>,NodeCompare> lista;
    set<Node*> elms; // another representation of "lista"

  public:
    hsExpandingQueue() {};
    bool empty() { return lista.empty(); };
    int size() { return lista.size(); };

    Node* pop() {
	Node* w = lista.top(); 
	lista.pop(); 
	elms.erase(w); 
	return w; 
    };

    void push(Node* w) { 
	if (elms.find(w) == elms.end()) {
	  lista.push(w); 
	  elms.insert(w); 
	}; 
    };
};

#endif
